import os
from pathlib import Path
from typing import Dict

import pytorch_lightning as pl
import pytorch_lightning.callbacks as pl_callbacks
import pytorch_lightning.loggers as pl_loggers
import pytorch_lightning.plugins.environments as pl_env_plugins
import torch
import torchmetrics
from pytorch_lightning.utilities import rank_zero_info

try:
    import rich  # noqa: F401

    HAS_RICH = True
except ImportError:
    HAS_RICH = False

import bcos.training.callbacks as custom_callbacks
from bcos.experiments.utils import CHECKPOINT_LAST_FILENAME, Experiment, sanitize_config
from bcos.training.agc import adaptive_clip_grad_
from bcos.training.ema import ExponentialMovingAverage


class ClassificationLitModel(pl.LightningModule):
    def __init__(self, dataset, base_network, experiment_name):
        super().__init__()
        self.experiment = Experiment(dataset, base_network, experiment_name)
        config = self.experiment.config
        model = self.experiment.get_model()

        self.save_hyperparameters()  # passed arguments
        self.save_hyperparameters(sanitize_config(config))  # the config as well

        self.experiment_name = experiment_name
        self.config = config
        self.model = model
        self.is_bcos = self.config["model"].get("is_bcos", False)
        self.criterion = self.config["criterion"]
        self.test_criterion = self.config["test_criterion"]

        num_classes = config["data"]["num_classes"]
        self.train_acc1 = torchmetrics.Accuracy(
            task="multiclass", top_k=1, num_classes=num_classes, compute_on_cpu=True
        )
        self.train_acc5 = torchmetrics.Accuracy(
            task="multiclass", top_k=5, num_classes=num_classes, compute_on_cpu=True
        )
        self.eval_acc1 = torchmetrics.Accuracy(
            task="multiclass", top_k=1, num_classes=num_classes, compute_on_cpu=True
        )
        self.eval_acc5 = torchmetrics.Accuracy(
            task="multiclass", top_k=5, num_classes=num_classes, compute_on_cpu=True
        )

        self.ema = None  # will be set during setup(stage="fit")
        self.ema_steps = None

        has_ema = self.config.get("ema", None) is not None
        self.eval_acc1_ema = self.eval_acc1.clone() if has_ema else None
        self.eval_acc5_ema = self.eval_acc5.clone() if has_ema else None

        self.lr_warmup_epochs = self.config["lr_scheduler"].warmup_epochs

        self.use_agc = self.config.get("use_agc", False)
        if self.use_agc:
            rank_zero_info("Adaptive Gradient Clipping is enabled!")

    def setup(self, stage: str) -> None:
        if stage != "fit":
            return

        ema_config = self.config.get("ema", None)
        if ema_config is None:
            return

        decay = ema_config["decay"]
        self.ema_steps = ema_config.get("steps", 32)
        rank_zero_info(f"Using EMA with {decay=} and steps={self.ema_steps}")

        # see https://github.com/pytorch/vision/blob/657c0767c5ca5564c8b437ac442/references/classification/train.py#L317
        adjust = (
            self.trainer.world_size
            * self.trainer.accumulate_grad_batches
            * self.config["data"]["batch_size"]
            * self.ema_steps
            / self.trainer.max_epochs
        )
        alpha = 1.0 - decay
        alpha = min(1.0, alpha * adjust)
        self.ema = ExponentialMovingAverage(self.model, decay=1.0 - alpha)
        self.ema.requires_grad_(False)

    def configure_optimizers(self):
        optimizer = self.config["optimizer"].create(self.model)
        scheduler = self.config["lr_scheduler"].create(
            optimizer,
            # this is total as in "whole" training
            total_steps=self.trainer.estimated_stepping_batches,
        )
        return dict(optimizer=optimizer, lr_scheduler=scheduler)

    def forward(self, in_tensor):
        return self.model(in_tensor)

    def training_step(self, batch, batch_idx):
        images, labels = batch
        outputs = self(images)

        loss = self.criterion(outputs, labels)
        with torch.no_grad():
            if labels.ndim == 2:
                # b/c of mixup/cutmix or sparse labels in general. See
                # https://github.com/pytorch/vision/blob/9851a69f6d294f5d672d973/references/classification/utils.py#L179
                labels = labels.argmax(dim=1)
            self.train_acc1(outputs, labels)
            self.train_acc5(outputs, labels)

            self.log("train_loss", loss)
            self.log(
                "train_acc1",
                self.train_acc1,
                on_step=True,
                on_epoch=True,
                prog_bar=True,
            )
            self.log(
                "train_acc5",
                self.train_acc5,
                on_step=True,
                on_epoch=True,
                prog_bar=True,
            )

            if self.ema is not None and batch_idx % self.ema_steps == 0:
                ema = self.ema
                ema.update_parameters(self.model)
                if self.trainer.current_epoch < self.lr_warmup_epochs:
                    ema.n_averaged.fill_(0)

        return loss

    def eval_step(self, batch, _batch_idx, val_or_test):
        images, labels = batch
        outputs = self(images)

        loss = self.test_criterion(outputs, labels)
        self.eval_acc1(outputs, labels)
        self.eval_acc5(outputs, labels)

        self.log(f"{val_or_test}_loss", loss)
        self.log(f"{val_or_test}_acc1", self.eval_acc1, on_epoch=True, prog_bar=True)
        self.log(f"{val_or_test}_acc5", self.eval_acc5, on_epoch=True, prog_bar=True)

        if self.ema is not None:
            ema = self.ema
            outputs = ema.module(images)
            ema_loss = self.test_criterion(outputs, labels)
            self.eval_acc1_ema(outputs, labels)
            self.eval_acc5_ema(outputs, labels)
            self.log(f"{val_or_test}_loss_ema", loss)
            self.log(f"{val_or_test}_acc1_ema", self.eval_acc1_ema, on_epoch=True)
            self.log(f"{val_or_test}_acc5_ema", self.eval_acc5_ema, on_epoch=True)
            return {"loss": loss, "loss_ema": ema_loss}
        else:
            return loss

    def validation_step(self, batch, batch_idx):
        return self.eval_step(batch, batch_idx, "val")

    def test_step(self, batch, batch_idx):
        return self.eval_step(batch, batch_idx, "test")

    def configure_gradient_clipping(
        self,
        optimizer,
        optimizer_idx,
        gradient_clip_val=None,
        gradient_clip_algorithm=None,
    ) -> None:
        # Note: this is called even if gradient_clip_val etc. is None
        if not self.use_agc:
            self.clip_gradients(optimizer, gradient_clip_val, gradient_clip_algorithm)
        else:
            adaptive_clip_grad_(self.parameters())

    def log_grad_norm(self, grad_norm_dict: Dict[str, float]) -> None:
        # we only care about total grad norm
        norm_type = float(self.trainer.track_grad_norm)
        total_norm = grad_norm_dict[f"grad_{norm_type}_norm_total"]
        del grad_norm_dict
        self.log(
            "gradients/total_norm",
            total_norm,
            on_step=True,
            on_epoch=True,
            prog_bar=False,
            logger=True,
        )

        if self.trainer.gradient_clip_val is not None:
            clipped_total_norm = min(
                float(self.trainer.gradient_clip_val), float(total_norm)
            )
            self.log(
                "gradients/clipped_total_norm",
                clipped_total_norm,
                on_step=True,
                on_epoch=True,
                prog_bar=False,
                logger=True,
            )


def put_trainer_args_into_trainer_config(args, trainer_config):
    if args.distributed:
        # https://github.com/Lightning-AI/lightning/discussions/6761#discussioncomment-2614296
        trainer_config["strategy"] = "ddp_find_unused_parameters_false"

    if args.fast_dev_run:
        trainer_config["fast_dev_run"] = True

    if hasattr(args, "nodes"):  # on slurm
        trainer_config["num_nodes"] = args.nodes

    if args.track_grad_norm:
        trainer_config["track_grad_norm"] = 2.0

    if hasattr(args, "amp") and args.amp:
        trainer_config["precision"] = 16

    if args.debug:
        trainer_config["deterministic"] = True


def run_training(args):
    """
    Instantiates everything and runs the training.
    """
    base_directory = args.base_directory
    dataset = args.dataset
    base_network = args.base_network
    experiment_name = args.experiment_name
    save_dir = Path(base_directory, dataset, base_network, experiment_name)
    save_dir.mkdir(parents=True, exist_ok=True)

    # set up loggers early so that WB starts capturing output asap
    loggers = setup_loggers(args)

    # get config
    exp = Experiment(dataset, base_network, experiment_name)
    config = exp.config.copy()

    # get and set seed
    seed = exp.config.get("seed", 42)
    pl.seed_everything(seed, workers=True)

    # init model
    model = ClassificationLitModel(
        dataset,
        base_network,
        experiment_name,
    )
    rank_zero_info(f"Model: {repr(model.model)}")

    # jit the internal model if specified
    if args.jit:
        model.model = torch.jit.script(model.model)
        rank_zero_info("Jitted the model!")

    # init datamodule
    datamodule = model.experiment.get_datamodule(
        cache_dataset=getattr(args, "cache_dataset", None),
    )

    # callbacks
    callbacks = setup_callbacks(args, config)

    # init trainer
    trainer_config = config["trainer"]
    put_trainer_args_into_trainer_config(args, trainer_config)

    # plugin for slurm
    if "SLURM_JOB_ID" in os.environ:  # we're on slurm
        # let submitit handle requeuing
        trainer_config["plugins"] = [
            pl_env_plugins.SLURMEnvironment(auto_requeue=False)
        ]

    trainer = pl.Trainer(
        default_root_dir=save_dir,
        accelerator="auto",
        devices="auto",
        logger=loggers,
        callbacks=callbacks,
        **trainer_config,
    )

    # decide whether to resume
    ckpt_path = None
    if args.resume:
        ckpt_path = save_dir / CHECKPOINT_LAST_FILENAME
        ckpt_path = ckpt_path if ckpt_path.exists() else None

    # start training
    trainer.fit(model, datamodule=datamodule, ckpt_path=ckpt_path)


def setup_loggers(args):
    loggers = []
    save_dir = Path(
        args.base_directory, args.dataset, args.base_network, args.experiment_name
    )

    if args.wandb_logger:
        wandb_logger = pl_loggers.WandbLogger(
            name=args.wandb_name or args.experiment_name,
            save_dir=str(save_dir),
            project=args.wandb_project,
            id=args.wandb_id,
        )
        loggers.append(wandb_logger)

    if args.csv_logger:
        csv_logger = pl_loggers.CSVLogger(
            save_dir=str(save_dir / "csv_logs"),
            name="",
            flush_logs_every_n_steps=1000,
        )
        loggers.append(csv_logger)

    if args.tensorboard_logger:
        tensorboard_logger = pl_loggers.TensorBoardLogger(
            save_dir=Path(
                "tb_logs",
                args.base_directory,
                args.dataset,
                args.base_network,
                args.experiment_name,
            ),
            name=args.experiment_name,
        )
        loggers.append(tensorboard_logger)

    return loggers


def setup_callbacks(args, config):
    callbacks = []
    save_dir = Path(
        args.base_directory, args.dataset, args.base_network, args.experiment_name
    )

    # the most important one
    save_callback = pl_callbacks.ModelCheckpoint(
        dirpath=save_dir,
        monitor="val_acc1",
        mode="max",
        filename="{epoch}-{val_acc1:.4f}",
        save_last=True,
        save_top_k=3,
        verbose=True,
    )
    callbacks.append(save_callback)

    use_ema = config.get("ema", None) is not None
    if use_ema:
        save_callback = pl_callbacks.ModelCheckpoint(
            dirpath=save_dir,
            monitor="val_acc1_ema",
            mode="max",
            filename="{epoch}-{val_acc1_ema:.4f}",
            save_top_k=3,
            verbose=True,
        )
        callbacks.append(save_callback)

    # lr monitor
    has_logger = args.wandb_logger or args.tensorboard_logger or args.csv_logger
    if has_logger:  # ow it's useless
        callbacks.append(pl_callbacks.LearningRateMonitor())
    slurm_or_submitit = hasattr(args, "nodes") or "SLURM_JOB_ID" in os.environ
    refresh_rate = args.refresh_rate or (20 if slurm_or_submitit else 5)
    if HAS_RICH and not slurm_or_submitit:
        callbacks.append(pl_callbacks.RichProgressBar(refresh_rate=refresh_rate))
    else:
        callbacks.append(pl_callbacks.TQDMProgressBar(refresh_rate=refresh_rate))

    # save metrics to checkpoint
    callbacks.append(custom_callbacks.MetricsTracker())

    # do explanation logging
    if args.explanation_logging:
        log_every = args.explanation_logging_every_n_epochs
        rank_zero_info(f"Will log explanations every {log_every} epoch(s)!")
        callbacks.append(
            custom_callbacks.ExplanationsLogger(log_every_n_epochs=log_every)
        )
    else:
        rank_zero_info("Explanation logging is disabled!")

    # for debugging purposes
    if args.debug:
        callbacks.append(custom_callbacks.ModelUpdateHasher())

    return callbacks



from pathlib import Path
from typing import Any, Dict

import numpy as np
import pytorch_lightning as pl
import pytorch_lightning.callbacks as pl_callbacks
import torchmetrics
from pytorch_lightning.utilities import rank_zero_only


class MetricsTracker(pl_callbacks.Callback):
    def __init__(self):
        self.metrics = None
        self.data = None

    @rank_zero_only
    def setup(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", stage: str
    ) -> None:
        self.metrics = {
            name: value
            for name, value in pl_module.named_children()
            if isinstance(value, torchmetrics.Metric)
        }
        self.data = {name: [] for name in self.metrics.keys()}

    @rank_zero_only
    def on_validation_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if trainer.sanity_checking:
            return
        for name in self.metrics.keys():
            value = float(self.metrics[name].compute())
            self.data[name].append((trainer.current_epoch, value))

    @rank_zero_only
    def on_save_checkpoint(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        checkpoint: Dict[str, Any],
    ) -> None:
        save_dir = Path(trainer.default_root_dir) / "metrics"
        save_dir.mkdir(exist_ok=True)

        # easier to load this way quickly for plotting etc.
        for name, values in self.data.items():
            np.savetxt(save_dir / f"{name}.gz", values)

    # easier to save this way to resume training etc.
    def state_dict(self):
        if self.data is None:
            return {}
        return self.data.copy()

    def load_state_dict(self, state_dict):
        if self.data is None:
            self.data = state_dict.copy()
        else:
            self.data.update(state_dict)



"""
B-cos VGG models

Modified from https://github.com/pytorch/vision/blob/0504df5ddf9431909130e7788faf05446bb8a2/torchvision/models/vgg.py
"""
import math
from typing import Any, Callable, Dict, List, Optional, Union, cast

import torch
import torch.nn as nn
import torch.nn.functional as F

from bcos.common import BcosUtilMixin
from bcos.modules import BcosConv2d, LogitLayer, norms

__all__ = [
    "BcosVGG",
    "vgg11",
    "vgg11_bnu",
    "vgg13",
    "vgg13_bnu",
    "vgg16",
    "vgg16_bnu",
    "vgg19_bnu",
    "vgg19",
]


DEFAULT_CONV_LAYER = BcosConv2d
DEFAULT_NORM_LAYER = norms.NoBias(norms.BatchNormUncentered2d)


class BcosVGG(BcosUtilMixin, nn.Module):
    def __init__(
        self,
        features: nn.Module,
        num_classes: int = 1000,
        init_weights: bool = True,
        conv_layer: Callable[..., nn.Module] = None,
        logit_bias: Optional[float] = None,
        logit_temperature: Optional[float] = None,
    ) -> None:
        super(BcosVGG, self).__init__()
        self.features = features
        # self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        self.classifier = nn.Sequential(
            conv_layer(512, 4096, kernel_size=7, padding=3, scale_fact=1000),
            # nn.ReLU(True),
            # nn.Dropout(),
            conv_layer(4096, 4096, scale_fact=1000),
            # nn.ReLU(True),
            # nn.Dropout(),
            conv_layer(4096, num_classes, scale_fact=1000),
        )
        self.num_classes = num_classes
        self.logit_layer = LogitLayer(
            logit_temperature=logit_temperature,
            logit_bias=logit_bias or -math.log(num_classes - 1),
        )
        if init_weights:
            self._initialize_weights()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.features(x)
        # x = self.avgpool(x)
        # x = torch.flatten(x, 1)[..., None, None]
        x = self.classifier(x)
        # because it's supposed to come after
        x = F.adaptive_avg_pool2d(x, (1, 1))
        x = torch.flatten(x, 1)
        x = self.logit_layer(x)
        return x

    def get_classifier(self) -> nn.Module:
        """Returns the classifier part of the model. Note this comes before global pooling."""
        return self.classifier

    def get_feature_extractor(self) -> nn.Module:
        """Returns the feature extractor part of the model. Without global pooling."""
        return self.features

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)


def make_layers(
    cfg: List[Union[str, int]],
    norm_layer: Callable[..., nn.Module] = DEFAULT_NORM_LAYER,
    conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
    in_channels: int = 6,
    no_pool=False,
) -> nn.Sequential:
    layers: List[nn.Module] = []
    new_config = []
    for idx, entry in enumerate(cfg):
        new_config.append([entry, 1])
        if entry == "M" and no_pool:
            new_config[idx - 1][1] = 2

    for v, stride in new_config:
        if v == "M":
            if no_pool:
                continue
            layers += [nn.AvgPool2d(kernel_size=2, stride=2)]
        else:
            v = cast(int, v)
            conv2d = conv_layer(
                in_channels, v, kernel_size=3, padding=1, stride=stride, scale_fact=1000
            )
            if not isinstance(norm_layer, nn.Identity):
                layers += [conv2d, norm_layer(v)]  # , nn.ReLU(inplace=True)]
            else:
                layers += [conv2d]
            in_channels = v
    return nn.Sequential(*layers)


cfgs: Dict[str, List[Union[str, int]]] = {
    "A": [64, "M", 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "B": [64, 64, "M", 128, 128, "M", 256, 256, "M", 512, 512, "M", 512, 512, "M"],
    "D": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        "M",
    ],
    "E": [
        64,
        64,
        "M",
        128,
        128,
        "M",
        256,
        256,
        256,
        256,
        "M",
        512,
        512,
        512,
        512,
        "M",
        512,
        512,
        512,
        512,
        "M",
    ],
}


def _vgg(
    arch: str,
    cfg: str,
    pretrained: bool,
    progress: bool,
    norm_layer: Callable[..., nn.Module] = None,
    conv_layer: Callable[..., nn.Module] = None,
    in_chans: int = 6,
    no_pool=False,
    **kwargs: Any
) -> BcosVGG:
    if pretrained:
        kwargs["init_weights"] = False
    model = BcosVGG(
        make_layers(
            cfgs[cfg],
            norm_layer=norm_layer,
            conv_layer=conv_layer,
            in_channels=in_chans,
            no_pool=no_pool,
        ),
        conv_layer=conv_layer,
        **kwargs,
    )
    if pretrained:
        raise ValueError(
            "If you want to load pretrained weights, then please use the entrypoints in "
            "bcos.pretrained or bcos.model.pretrained instead."
        )
    return model


def vgg11(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> BcosVGG:
    return _vgg("vgg11", "A", pretrained, progress, norm_layer=nn.Identity, **kwargs)


def vgg11_bnu(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosVGG:
    return _vgg("vgg11_bnu", "A", pretrained, progress, **kwargs)


def vgg13(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> BcosVGG:
    return _vgg("vgg13", "B", pretrained, progress, norm_layer=nn.Identity, **kwargs)


def vgg13_bnu(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosVGG:
    return _vgg("vgg13_bnu", "B", pretrained, progress, **kwargs)


def vgg16(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> BcosVGG:
    return _vgg("vgg16", "D", pretrained, progress, norm_layer=nn.Identity, **kwargs)


def vgg16_bnu(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosVGG:
    return _vgg("vgg16_bnu", "D", pretrained, progress, **kwargs)


def vgg19(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> BcosVGG:
    return _vgg("vgg19", "E", pretrained, progress, norm_layer=nn.Identity, **kwargs)


def vgg19_bnu(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosVGG:
    return _vgg("vgg19_bnu", "E", pretrained, progress, **kwargs)



import copy
import io
import math
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import pytorch_lightning.callbacks as pl_callbacks
import pytorch_lightning.loggers as pl_loggers
import torch
import torchvision.transforms.functional as transformsF
from mpl_toolkits.axes_grid1 import ImageGrid
from PIL import Image
from pytorch_lightning.utilities import rank_zero_only, rank_zero_warn

from bcos.common import gradient_to_image


def to_numpy(tensor: "Union[torch.Tensor, np.ndarray]") -> "np.ndarray":
    if not isinstance(tensor, torch.Tensor):
        return tensor
    return tensor.detach().cpu().numpy()


# class that creates a cpu copy of the model
class ModelCopy(torch.nn.Module):
    def __init__(self, model: "torch.nn.Module", use_cpu: bool = True):
        super().__init__()
        self.device = (
            torch.device("cpu") if use_cpu else next(model.parameters()).device
        )
        self.model = copy.deepcopy(model).to(self.device)
        self.original_model = model  # ref to original model

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def explanation_mode(self, *args, **kwargs):
        return self.model.explanation_mode(*args, **kwargs)

    def update(self):
        """Update the copy with the original model's parameters."""
        with torch.no_grad():
            # get original model's state dict and put tensors on given device
            state_dict = {
                k: v.to(self.device, non_blocking=True, copy=True)
                for k, v in self.original_model.state_dict().items()
            }
            self.model.load_state_dict(state_dict)

    def freeze(self):
        for p in self.model.parameters():
            p.requires_grad = False


# explanation logger
class ExplanationsLogger(pl_callbacks.Callback):
    def __init__(
        self,
        log_every_n_epochs: int = 1,
        idx: Optional[torch.Tensor] = None,
        max_imgs: int = 32,
    ):
        self.log_every_n_epochs = log_every_n_epochs
        self.max_imgs = max_imgs
        self.idx = idx

        self.imgs = None
        self.lbls = None

        self.has_bcos_model = False

        self.model: "Optional[ModelCopy]" = None

    @rank_zero_only
    def setup(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", stage: str
    ) -> None:
        if stage != "fit":
            return

        if hasattr(trainer, "datamodule") and trainer.datamodule is not None:
            dataset = trainer.datamodule.eval_dataset
        else:
            dataset = trainer.val_dataloaders[0].dataset

        # if not bcos specified in pl_module
        # *and* if it neither inherits from BcosModelBase; do not do expl logging
        if not getattr(pl_module, "is_bcos", False):
            rank_zero_warn(
                "Explanation logger is active but PLModule doesn't specify a B-cos model! "
                "Explanation logging will be skipped."
            )
            return

        # we have a bcos model so we now we can safely do explanation log
        self.has_bcos_model = True

        max_imgs = self.max_imgs
        idx = self.idx
        if idx is None:
            g = torch.Generator().manual_seed(42)
            idx = torch.randint(high=len(dataset), size=(max_imgs,), generator=g)
        self.idx = idx[:max_imgs]

        imgs = []
        lbls = []
        for i in idx:
            img, lbl = dataset[i]
            imgs.append(img)
            lbls.append(lbl)
        self.imgs = torch.stack(imgs)
        self.lbls = torch.tensor(lbls)

        if self.has_bcos_model and not hasattr(pl_module.model, "explanation_mode"):
            raise RuntimeError(
                "PLModule specified a bcos model. "
                "However, model doesn't contain .explanation_mode()!"
            )
        self.model = ModelCopy(pl_module.model, use_cpu=True)

    @rank_zero_only
    def on_validation_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if not self.has_bcos_model:
            return  # no point in logging

        if trainer.sanity_checking:
            return

        if (trainer.current_epoch + 1) % self.log_every_n_epochs != 0:
            return

        # not initialized yet
        if self.model is None or self.imgs is None or self.lbls is None:
            return

        wandb_logger = None
        tensorboard_logger = None
        for logger in trainer.loggers:
            if isinstance(logger, pl_loggers.WandbLogger):
                wandb_logger = logger
            elif isinstance(logger, pl_loggers.TensorBoardLogger):
                tensorboard_logger = logger

        self.log_explanations(wandb_logger, tensorboard_logger)

    def log_explanations(
        self,
        wandb_logger: "Optional[pl_loggers.WandbLogger]" = None,
        tensorboard_logger: "Optional[pl_loggers.TensorBoardLogger]" = None,
        postfix="",
    ):
        if wandb_logger is None and tensorboard_logger is None:
            return

        results = self.get_results()
        explanations = self.plot_explanations(results)
        expl_pil_img = self.fig_to_pil(explanations)

        namespace = f"explanations{postfix}"
        if wandb_logger:
            wandb_logger.log_image(f"Explanations/{namespace}", [expl_pil_img])

        if tensorboard_logger:
            writer = tensorboard_logger.experiment
            if hasattr(writer, "add_figure"):
                writer.add_figure(f"{namespace}/Explanations", explanations)
            else:
                writer.add_image(
                    f"{namespace}/Explanations", transformsF.to_tensor(expl_pil_img)
                )

        # figures are retained, hence manual clean up
        plt.close(explanations)

    def get_results(self):
        # prepare things
        imgs = self.imgs
        lbls = self.lbls
        assert self.model is not None
        self.model.update()
        self.model.eval()
        self.model.freeze()  # just to be sure
        device = self.model.device

        # result dict
        results = []  # (img, lbl, pred, w)
        with torch.enable_grad(), self.model.explanation_mode():
            # calculate expls
            for img, lbl in zip(imgs, lbls):
                assert not img.requires_grad and img.grad is None
                # prep data
                img = img[None].to(device).requires_grad_()
                lbl = lbl.item()

                # get predictions
                out = self.model(img)
                pred = out.max(1)
                pred.values.backward(inputs=[img])
                pred = pred.indices.item()

                # dynamic weights
                grad = img.grad

                # add to results
                results.append(
                    (
                        img.detach().cpu()[0],
                        lbl,
                        pred,
                        grad.detach().cpu()[0],
                    )
                )

        return results

    def plot_explanations(self, results):
        N = len(results) * 2
        H, W = self.get_grid_size(N)

        fig = plt.figure(dpi=200)
        grid = ImageGrid(
            fig,
            111,  # similar to subplot(111)
            nrows_ncols=(H, W),
            axes_pad=0.0,  # is okay
        )

        for i, (img, lbl, pred, weight) in enumerate(results):
            x = i // W
            y = i - W * x
            img_ax = grid[2 * x * W + y]
            expl_ax = grid[(2 * x + 1) * W + y]

            expl = gradient_to_image(img, weight)

            img_ax.imshow(to_numpy(img[:3].permute(1, 2, 0)))
            expl_ax.imshow(to_numpy(expl))

            # style
            for ax in (img_ax, expl_ax):
                ax.set_xticks([])
                ax.set_yticks([])
            for spine in expl_ax.spines.values():
                spine.set_edgecolor("lime" if lbl == pred else "orange")
                spine.set_linestyle((0, (5, 3)) if lbl == pred else (0, (2, 3)))

        return fig

    @staticmethod
    def get_grid_size(n, ratio=2 / 3):
        """Returns grid size (H,W) with
        H/W approx= ratio and H*W>=n. and H=even.
        """
        x = int(math.sqrt(ratio * n))
        if x % 2 == 1:
            x += 1
        # doesn't have to be perfect just good enough
        y = int(math.ceil(n / x))
        return x, y

    @staticmethod
    def fig_to_pil(fig):
        # from https://stackoverflow.com/a/61754995/10614892
        buf = io.BytesIO()
        fig.savefig(buf, format="jpeg", bbox_inches="tight", pad_inches=0)
        buf.seek(0)
        img = Image.open(buf)
        return img

    @staticmethod
    def convert_image(pil_img):
        # TODO: if needed change
        return pil_img



import warnings
from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch.linalg as LA
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from .common import DetachableModule

__all__ = ["NormedConv2d", "BcosConv2d", "BcosConv2dWithScale"]


class NormedConv2d(nn.Conv2d):
    """
    Standard 2D convolution, but with unit norm weights.
    """

    def forward(self, in_tensor: Tensor) -> Tensor:
        w = self.weight / LA.vector_norm(self.weight, dim=(1, 2, 3), keepdim=True)
        return self._conv_forward(input=in_tensor, weight=w, bias=self.bias)


class BcosConv2d(DetachableModule):
    """
    BcosConv2d is a 2D convolution with unit norm weights and a cosine similarity
    activation function. The cosine similarity is calculated between the
    convolutional patch and the weight vector. The output is then scaled by the
    cosine similarity.

    See the paper for more details: https://arxiv.org/abs/2205.10268

    Parameters
    ----------
    in_channels : int
        Number of channels in the input image
    out_channels : int
        Number of channels produced by the convolution
    kernel_size : int | tuple[int, ...]
        Size of the convolving kernel
    stride : int | tuple[int, ...]
        Stride of the convolution. Default: 1
    padding : int | tuple[int, ...]
        Zero-padding added to both sides of the input. Default: 0
    dilation : int | tuple[int, ...]
        Spacing between kernel elements. Default: 1
    groups : int
        Number of blocked connections from input channels to output channels.
        Default: 1
    padding_mode : str
        Padding mode. One of ``'zeros'``, ``'reflect'``, ``'replicate'`` or ``'circular'``.
        Default: ``'zeros'``
    device : Optional[torch.device]
        The device of the weights.
    dtype : Optional[torch.dtype]
        The dtype of the weights.
    b : int | float
        The base of the exponential used to scale the cosine similarity.
    max_out : int
        Number of MaxOut units to use. If 1, no MaxOut is used.
    **kwargs : Any
        Ignored.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, ...]] = 1,
        stride: Union[int, Tuple[int, ...]] = 1,
        padding: Union[int, Tuple[int, ...]] = 0,
        dilation: Union[int, Tuple[int, ...]] = 1,
        groups: int = 1,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
        # special (note no scale here! See BcosConv2dWithScale below)
        b: Union[int, float] = 2,
        max_out: int = 1,
        **kwargs,  # bias is always False
    ):
        assert max_out > 0, f"max_out should be greater than 0, was {max_out}"
        super().__init__()

        # save everything
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size

        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.padding_mode = padding_mode
        self.device = device
        self.dtype = dtype
        self.bias = False

        self.b = b
        self.max_out = max_out

        # check dilation
        if dilation > 1:
            warnings.warn("dilation > 1 is much slower!")
            self.calc_patch_norms = self._calc_patch_norms_slow

        self.linear = NormedConv2d(
            in_channels=in_channels,
            out_channels=out_channels * max_out,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=False,
            padding_mode=padding_mode,
            device=device,
            dtype=dtype,
        )

    def forward(self, in_tensor: Tensor) -> Tensor:
        """
        Forward pass implementation.
        Args:
            in_tensor: Input tensor. Expected shape: (B, C, H, W)

        Returns:
            BcosConv2d output on the input tensor.
        """
        return self.forward_impl(in_tensor)

    def forward_impl(self, in_tensor: Tensor) -> Tensor:
        """
        Forward pass.
        Args:
            in_tensor: Input tensor. Expected shape: (B, C, H, W)

        Returns:
            BcosConv2d output on the input tensor.
        """
        # Simple linear layer
        out = self.linear(in_tensor)

        # MaxOut computation
        if self.max_out > 1:
            M = self.max_out
            O = self.out_channels  # noqa: E741
            out = out.unflatten(dim=1, sizes=(O, M))
            out = out.max(dim=2, keepdim=False).values

        # if B=1, no further calculation necessary
        if self.b == 1:
            return out

        # Calculating the norm of input patches: ||x||
        norm = self.calc_patch_norms(in_tensor)

        # Calculate the dynamic scale (|cos|^(B-1))
        # Note that cos = (x·ŵ)/||x||
        maybe_detached_out = out
        if self.detach:
            maybe_detached_out = out.detach()
            norm = norm.detach()

        if self.b == 2:
            dynamic_scaling = maybe_detached_out.abs() / norm
        else:
            abs_cos = (maybe_detached_out / norm).abs() + 1e-6
            dynamic_scaling = abs_cos.pow(self.b - 1)

        # put everything together
        out = dynamic_scaling * out  # |cos|^(B-1) (ŵ·x)
        return out

    def calc_patch_norms(self, in_tensor: Tensor) -> Tensor:
        """
        Calculates the norms of the patches.
        """
        squares = in_tensor**2
        if self.groups == 1:
            # normal conv
            squares = squares.sum(1, keepdim=True)
        else:
            G = self.groups
            C = self.in_channels
            # group channels together and sum reduce over them
            # ie [N,C,H,W] -> [N,G,C//G,H,W] -> [N,G,H,W]
            # note groups MUST come first
            squares = squares.unflatten(1, (G, C // G)).sum(2)

        norms = (
            F.avg_pool2d(
                squares,
                self.kernel_size,
                padding=self.padding,
                stride=self.stride,
                divisor_override=1,
            )
            + 1e-6  # stabilizing term
        ).sqrt_()

        if self.groups > 1:
            # norms.shape will be [N,G,H,W] (here H,W are spatial dims of output)
            # we need to convert this into [N,O,H,W] so that we can divide by this norm
            # (because we can't easily do broadcasting)
            N, G, H, W = norms.shape
            O = self.out_channels  # noqa: E741
            norms = torch.repeat_interleave(norms, repeats=O // G, dim=1)

        return norms

    def _calc_patch_norms_slow(self, in_tensor: Tensor) -> Tensor:
        # this is much slower but definitely correct
        # use for testing or something difficult to implement
        # like dilation
        ones_kernel = torch.ones_like(self.linear.weight)

        return (
            F.conv2d(
                in_tensor**2,
                ones_kernel,
                None,
                self.stride,
                self.padding,
                self.dilation,
                self.groups,
            )
            + 1e-6
        ).sqrt_()

    def extra_repr(self) -> str:
        # rest in self.linear
        s = "B={b}"

        if self.max_out > 1:
            s += ", max_out={max_out}"

        # final comma as self.linear is shown in next line
        s += ","

        return s.format(**self.__dict__)


class BcosConv2dWithScale(BcosConv2d):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, ...]] = 1,
        stride: Union[int, Tuple[int, ...]] = 1,
        padding: Union[int, Tuple[int, ...]] = 0,
        dilation: Union[int, Tuple[int, ...]] = 1,
        groups: int = 1,
        padding_mode: str = "zeros",
        device=None,
        dtype=None,
        # special
        b: Union[int, float] = 2,
        max_out: int = 1,
        # the following come from the original v1 https://github.com/moboehle/B-cos
        scale: Optional[float] = None,  # use default init
        scale_factor: Union[int, float] = 100.0,
        **kwargs,
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            padding_mode,
            device,
            dtype,
            b,
            max_out,
            **kwargs,
        )

        if scale is None:
            ks_scale = (
                kernel_size
                if not isinstance(kernel_size, tuple)
                else np.sqrt(np.prod(kernel_size))
            )
            self.scale = (ks_scale * np.sqrt(self.in_channels)) / scale_factor
        else:
            assert scale != 1.0, "For scale=1.0, use the normal BcosConv2d instead!"
            self.scale = scale

        warnings.warn(
            "BcosConv2dWithScale is deprecated and will be removed in a future version. "
            "Use BcosConv2d with scale=1.0 instead.",
            DeprecationWarning,
        )

    def forward(self, in_tensor: Tensor) -> Tensor:
        out = self.forward_impl(in_tensor)
        return out / self.scale

    def extra_repr(self) -> str:
        result = super().extra_repr()
        result = f"scale={self.scale:.3f}, " + result
        return result



"""
Constants based on the structure of the config and results storing system.
"""
# these are the ones you probably want to change if then
ROOT = "bcos"
BASE_EXPERIMENTS_DIRECTORY = "experiments"

# you probably don't want to change these
MODEL_FACTORY_MODULE = "model"
MODEL_FACTORY_VAR_NAME = "get_model"
CONFIGS_MODULE = "experiment_parameters"
CONFIGS_VAR_NAME = "CONFIGS"


CHECKPOINT_LAST_FILENAME = "last.ckpt"



"""
Utilities for loading models, configs, etc.
"""
import difflib
import inspect
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import torch

from ..config_utils import get_configs_and_model_factory, update_config
from ..exceptions import ModelFactoryDoesNotSupportPretrainedError
from ..structure_constants import BASE_EXPERIMENTS_DIRECTORY, ROOT
from .loading_utils import (
    device_safe_load_state_dict_from_path,
    get_last_checkpoint_path_in_save_dir,
    get_state_dict_and_training_ckpt_from_save_dir,
)
from .metric_utils import Metrics

__all__ = [
    "Experiment",  # main API
]


class Experiment:
    """
    Utility class for loading models and configs.

    Examples
    --------
    >>> from bcos.experiments.utils import Experiment
    >>> exp = Experiment("ImageNet", "bcos_final", "resnet_50")
    >>> exp.save_dir                 # gets the save directory
    >>> exp.config                   # gets the config
    >>> exp.get_model()              # gets a new model instance
    >>> exp.available_checkpoints()  # gets the available checkpoints
    >>> exp.get_datamodule()         # gets a new datamodule instance (requires PL)
    >>> exp.load_metrics()           # loads the metrics
    >>> exp.load_trained_model()     # creates a model and loads the weights from training ckpt
    """

    def __init__(
        self,
        path_or_dataset: Union[str, Path],
        base_network: Optional[str] = None,
        experiment_name: Optional[str] = None,
        base_directory: str = BASE_EXPERIMENTS_DIRECTORY,
    ):
        """
        Initialize an Experiment object.

        Parameters
        ----------
        path_or_dataset : Union[str, Path]
            The path to a stored experiment OR the name of the dataset.
        base_network : Optional[str]
            The base network. Only provide this if `path_or_dataset` is a dataset.
        experiment_name : Optional[str]
            The experiment name. Only provide this if `path_or_dataset` is a dataset.
        base_directory : str
            The base directory. Default: "./experiments"
        """
        if (
            isinstance(path_or_dataset, Path) or "/" in path_or_dataset
        ):  # with / will assume it's a path
            assert (
                base_network is None and experiment_name is None
            ), f"Path provided ('{path_or_dataset=}' b/c it contains '/')! Other arguments should be None!"
            path = Path(path_or_dataset)
            experiment_name = path.name
            base_network = path.parent.name
            dataset = path.parent.parent.name
            base_directory = path.parent.parent.parent
        else:  # assume it's a dataset
            dataset = path_or_dataset
            assert (
                base_network is not None and experiment_name is not None
            ), "Provide other arguments too!"

        configs, get_model = get_configs_and_model_factory(dataset, base_network)

        self._get_model = get_model
        """ The model factory. """

        if experiment_name not in configs:
            msg = f"Config for '{experiment_name=}' not found!"
            maybe = difflib.get_close_matches(experiment_name, configs.keys())
            if maybe:
                msg += f" Did you mean '{maybe[0]}'?"
            raise ValueError(msg)

        self.config = configs[experiment_name]
        self.base_directory = base_directory
        self.dataset = dataset
        self.base_network = base_network
        self.experiment_name = experiment_name
        self.save_dir = Path(base_directory, dataset, base_network, experiment_name)

    def get_datamodule(self, **data_config_overrides):
        """The PytorchLightning datamodule for the experiment.
        Need to set up before you can use. E.g.:

        Example
        -------
        >>> exp = Experiment(...)
        >>> datamodule = exp.get_datamodule()
        >>> datamodule.setup("val")
        >>> val_dataloader = datamodule.val_dataloader()
        """
        # not a hard dependency
        from bcos.data.datamodules import ClassificationDataModule

        data_config = self.config["data"]
        data_config.update(data_config_overrides)
        registry = ClassificationDataModule.registry()
        if self.dataset in registry:
            datamodule = registry[self.dataset](data_config)
        else:
            available = list(registry.keys())
            _line_break = "\n" + " " * len("ValueError: ")
            raise ValueError(
                f"Dataset '{self.dataset}' not found! Available: {available}{_line_break}"
                f"If you're adding a new dataset, subclass ClassificationDataModule{_line_break}"
                f"in bcos.data.datamodules and make sure it ends with 'DataModule'{_line_break}"
                "(e.g. 'MyDatasetDataModule') (the suffix will be stripped)."
            )

        return datamodule

    @property
    def last_training_checkpoint_path(self) -> Path:
        """
        Get the path of the last checkpoint for this experiment.
        """
        return get_last_checkpoint_path_in_save_dir(self.save_dir)

    @property
    def last_training_checkpoint_state_dict(self) -> Dict[str, Any]:
        """
        Try to load the last checkpoint state dict.
        """
        save_path = self.last_training_checkpoint_path
        return device_safe_load_state_dict_from_path(save_path)

    def get_model(self, **kwargs: Any):
        """
        Create a new instance of the model for the experiment.

        Parameters
        ----------
        **kwargs
            Any additional keyword arguments to pass to the model constructor.

        Returns
        -------
        torch.nn.Module
            The model.
        """
        model_config = self.config["model"]
        model_config = update_config(model_config, kwargs)
        return self._get_model(model_config)

    def load_metrics(self) -> Metrics:
        """
        Returns
        -------
        Metrics
            Load a dictionary containing metrics from the last state dict.
            Tensor shape: [num_epochs (+1), 2] with columns (#epoch, metric_value)
            You can use the `.best_epoch_and_accuracy` attribute of the returned
            dict to get tuple containing the best epoch and accuracy. (See also
            `.get_best_epoch_and_metric_value_for(key)` method.)
        """
        return Metrics.from_experiment_dir(self.save_dir)

    def load_trained_model(
        self,
        reload: str = "last",
        ema: bool = False,
        verbose: bool = False,
        return_training_ckpt_if_possible: bool = False,
        **kwargs: Any,
    ) -> Union[torch.nn.Module, Dict[str, Union[torch.nn.Module, Dict[str, Any]]]]:
        """
        Load a trained model.

        Parameters
        ----------
        reload : str
            What weights to reload. Either "last", "best", "best_any" or "epoch_<N>".
            "best_any": will check both non-ema or ema (if they exist) weights for best accuracy
            "best" will load best ema weights if `ema=True` and best non-ema otherwise
            Default: "last"
        ema : bool
            Only consider EMA weights. Default: False
            Not applicable if `reload="best_any"`.
        verbose : bool
            Print information about loaded epoch and top-1 accuracy.
            Default: False
        return_training_ckpt_if_possible: bool
            Whether to ALSO return the whole training checkpoint dict instead.
            If `True`, then the return type will always be a dictionary with keys
            "model" and "ckpt" (for model and checkpoint respectively).
            Default: False
        **kwargs: Any
            Any additional keyword arguments to override the model config.

        Returns
        -------
        torch.nn.Module | Dict[str, torch.nn.Module | Dict[str, Any]]
            A trained model.
            Or a dictionary containing the model and the training checkpoint dict
            (with keys "model" and "ckpt" respectively).
        """
        model = self.get_model(**kwargs)

        # check if we have training checkpoints
        if (
            not self.save_dir.exists()
            or len(self.available_checkpoints(suppress_warnings=True)) == 0
        ):
            if verbose:
                print("No checkpoints found! Trying to load external weights...")
            if return_training_ckpt_if_possible:
                warnings.warn("No checkpoints found! Returning model only.")
            try:
                # if no training checkpoints available, then we try to load external weights
                model = self._try_load_external_weights(verbose, reload=reload, ema=ema)
                return (
                    model
                    if not return_training_ckpt_if_possible
                    else dict(
                        model=model,
                        ckpt=None,
                    )
                )
            except ModelFactoryDoesNotSupportPretrainedError:
                raise FileNotFoundError(
                    "No checkpoints found (and no external weights either)!"
                )

        # have training checkpoints, so load them
        state_dict, training_ckpt = get_state_dict_and_training_ckpt_from_save_dir(
            self.save_dir,
            reload=reload,
            ema=ema,
            verbose=verbose,
        )
        model.load_state_dict(state_dict)

        if return_training_ckpt_if_possible:
            return dict(model=model, ckpt=training_ckpt)
        else:
            return model

    def available_checkpoints(self, suppress_warnings: bool = False) -> List[Path]:
        """
        Returns a list to paths to available checkpoint files if any.

        Parameters
        ----------
        suppress_warnings : bool
            Whether to suppress warnings if no checkpoints are found.
            Default: False

        Returns
        -------
        List[Path]
            A list of paths to available checkpoint files.
        """
        if not self.save_dir.exists():
            raise ValueError(f"Save directory '{self.save_dir}' does not exist!")

        result = list(p for p in self.save_dir.iterdir() if p.suffix == ".ckpt")

        if len(result) == 0 and not suppress_warnings:
            warnings.warn(f"No checkpoints found in '{self.save_dir}'!")
        return result

    def _try_load_external_weights(
        self, verbose: bool, **kwargs: Any
    ) -> torch.nn.Module:
        """
        Try to load external weights via model factory.

        Parameters
        ----------
        verbose : bool
            Whether to print information about loaded epoch and top-1 accuracy.
        kwargs : Any
            Keyword arguments to check for `ema` and `reload` (which will be ignored)
            to inform the user about it.

        Returns
        -------
        torch.nn.Module
            Model loaded with (external) weights from the model factory.
        """
        if kwargs.get("ema", False):
            warnings.warn("ema=True will be ignored for external weights!")
        if kwargs.get("reload", "last") != "last":
            warnings.warn(
                "Seems like you're using a non-default value of reload. "
                "This will be ignored for external weights!"
            )

        # use inspect to check if the model factory has a `pretrained` argument
        # if it does, then we pass it to the model factory
        # otherwise, we just pass it to the model constructor
        model_factory = self._get_model
        model_factory_signature = inspect.signature(model_factory)
        if "pretrained" in model_factory_signature.parameters:
            model_config = self.config["model"]
            model = model_factory(model_config, pretrained=True)
        else:
            if verbose:
                print("No pretrained argument found in model factory!")
            raise ModelFactoryDoesNotSupportPretrainedError

        if verbose:
            print("Loaded external weights!")
        return model



"""
Convert a dataset into a in-memory cached dataset.
We use https://tutorials.baguasys.com/more-optimizations/cached-dataset

We only store loaded images to prevent high IO.
From https://github.com/BaguaSys/bagua
"""
import torch.utils.data as data
from torchvision.datasets import ImageFolder

from .cached_loader import CacheLoader

DEFAULT_CAPACITY = 200 * (1024**3)


class CachedImageFolder(data.Dataset):
    def __init__(self, underlying: ImageFolder, capacity: int = DEFAULT_CAPACITY):
        super().__init__()

        self.underlying = underlying
        self.transform = self.underlying.transform
        self.target_transform = self.underlying.target_transform
        self.samples = self.underlying.samples
        self.loader = self.underlying.loader

        self.cache_loader = CacheLoader(
            backend="redis",
            capacity_per_node=capacity,
            hosts=None,
            cluster_mode=True,
        )

    def __len__(self):
        return len(self.underlying)

    def __getitem__(self, index):
        # based on ImageFolder.__getitem__
        path, target = self.samples[index]
        sample = self.cache_loader.get(path, self.loader)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return sample, target



from functools import partial

import bcos.models.resnet
from bcos.modules.bcosconv2d import BcosConv2d, BcosConv2dWithScale

__all__ = ["get_model"]


def get_model(model_config):
    assert model_config.get("is_bcos", False), "Should be true!"
    # extract args
    arch_name = model_config["name"]
    args = model_config["args"]
    bcos_args = model_config["bcos_args"]

    # specify conv layer
    # unlike norm_layer these stay the same class
    if model_config.get("use_bcos_scale", False):
        args["conv_layer"] = partial(BcosConv2dWithScale, **bcos_args)
    else:
        args["conv_layer"] = partial(BcosConv2d, **bcos_args)

    # create model
    model = getattr(bcos.models.resnet, f"cifar10_{arch_name}")(**args)

    return model



from typing import Optional

import torch.nn as nn
from torch import Tensor

__all__ = [
    "LogitLayer",
]


class LogitLayer(nn.Module):
    def __init__(
        self,
        logit_temperature: Optional[float] = None,
        logit_bias: Optional[float] = None,
    ):
        # note: T=None => T=1 and b=None => b=0
        super().__init__()
        self.logit_bias = logit_bias
        self.logit_temperature = logit_temperature

    def forward(self, in_tensor: Tensor) -> Tensor:
        if self.logit_temperature is not None:
            in_tensor = in_tensor / self.logit_temperature
        if self.logit_bias is not None:
            in_tensor = in_tensor + self.logit_bias
        return in_tensor

    def extra_repr(self) -> str:
        ret = ""
        if self.logit_temperature is not None:
            ret += f"logit_temperature={self.logit_temperature}, "
        if self.logit_bias is not None:
            ret += f"logit_bias={self.logit_bias}, "
        ret = ret[:-2]
        return ret



"""
B-cos ConvNeXt models

Modified from https://github.com/pytorch/vision/blob/0504df5ddf9431909130e7788faf05446bb/torchvision/models/convnext.py
"""
import math
from functools import partial
from typing import Any, Callable, List, Optional, Sequence

import torch
from torch import Tensor, nn
from torchvision.ops import StochasticDepth

from bcos.common import BcosUtilMixin
from bcos.modules import BcosConv2d, LogitLayer, norms

__all__ = [
    "BcosConvNeXt",
    "convnext_atto",
    "convnext_tiny",
    "convnext_small",
    "convnext_base",
    "convnext_large",
    "CNBlock",
    "CNBlockConfig",
]

DEFAULT_NORM_LAYER = norms.NoBias(norms.DetachablePositionNorm2d)
DEFAULT_CONV_LAYER = BcosConv2d


class CNBlock(nn.Module):
    def __init__(
        self,
        dim,
        layer_scale: float,
        stochastic_depth_prob: float,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
        norm_layer: Optional[Callable[..., nn.Module]] = DEFAULT_NORM_LAYER,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = partial(DEFAULT_NORM_LAYER, eps=1e-6)

        self.block = nn.Sequential(
            conv_layer(dim, dim, kernel_size=7, padding=3, groups=dim, bias=False),
            # Permute([0, 2, 3, 1]),
            norm_layer(dim),
            conv_layer(
                in_channels=dim, out_channels=4 * dim, kernel_size=1, bias=False
            ),
            # nn.GELU(),  # The B-cos transform itself is non-linear
            conv_layer(
                in_channels=4 * dim, out_channels=dim, kernel_size=1, bias=False
            ),
            # Permute([0, 3, 1, 2]),
        )
        self.layer_scale = nn.Parameter(torch.ones(dim, 1, 1) * layer_scale)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")

    def forward(self, input: Tensor) -> Tensor:
        result = self.layer_scale * self.block(input)
        result = self.stochastic_depth(result)
        result += input
        return result


class CNBlockConfig:
    # Stores information listed at Section 3 of the ConvNeXt paper
    def __init__(
        self,
        input_channels: int,
        out_channels: Optional[int],
        num_layers: int,
    ) -> None:
        self.input_channels = input_channels
        self.out_channels = out_channels
        self.num_layers = num_layers

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        s += "input_channels={input_channels}"
        s += ", out_channels={out_channels}"
        s += ", num_layers={num_layers}"
        s += ")"
        return s.format(**self.__dict__)


class BcosConvNeXt(BcosUtilMixin, nn.Module):
    def __init__(
        self,
        block_setting: List[CNBlockConfig],
        stochastic_depth_prob: float = 0.0,
        layer_scale: float = 1e-6,
        num_classes: int = 1000,
        in_chans: int = 6,
        block: Optional[Callable[..., nn.Module]] = None,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        logit_bias: Optional[float] = None,
        logit_temperature: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        # _log_api_usage_once(self)

        if not block_setting:
            raise ValueError("The block_setting should not be empty")
        elif not (
            isinstance(block_setting, Sequence)
            and all([isinstance(s, CNBlockConfig) for s in block_setting])
        ):
            raise TypeError("The block_setting should be List[CNBlockConfig]")

        if block is None:
            block = CNBlock

        if norm_layer is None:
            norm_layer = partial(DEFAULT_NORM_LAYER, eps=1e-6)

        layers: List[nn.Module] = []

        # Stem
        firstconv_output_channels = block_setting[0].input_channels
        layers.extend(
            [
                conv_layer(
                    in_chans,
                    firstconv_output_channels,
                    kernel_size=4,
                    stride=4,
                    padding=0,
                    bias=False,
                ),
                norm_layer(firstconv_output_channels),
            ]
        )

        total_stage_blocks = sum(cnf.num_layers for cnf in block_setting)
        stage_block_id = 0
        for cnf in block_setting:
            # Bottlenecks
            stage: List[nn.Module] = []
            for _ in range(cnf.num_layers):
                # adjust stochastic depth probability based on the depth of the stage block
                sd_prob = (
                    stochastic_depth_prob * stage_block_id / (total_stage_blocks - 1.0)
                )
                stage.append(
                    block(
                        cnf.input_channels,
                        layer_scale,
                        sd_prob,
                        conv_layer=conv_layer,
                        norm_layer=norm_layer,
                    )
                )
                stage_block_id += 1
            layers.append(nn.Sequential(*stage))
            if cnf.out_channels is not None:
                # Downsampling
                layers.append(
                    nn.Sequential(
                        norm_layer(cnf.input_channels),
                        conv_layer(
                            cnf.input_channels,
                            cnf.out_channels,
                            kernel_size=2,
                            stride=2,
                            bias=False,
                        ),
                    )
                )

        self.features = nn.Sequential(*layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)

        lastblock = block_setting[-1]
        lastconv_output_channels = (
            lastblock.out_channels
            if lastblock.out_channels is not None
            else lastblock.input_channels
        )
        self.classifier = nn.Sequential(
            norm_layer(lastconv_output_channels),
            conv_layer(
                lastconv_output_channels, num_classes, kernel_size=1, bias=False
            ),
        )
        self.num_classes = num_classes
        self.logit_layer = LogitLayer(
            logit_temperature=logit_temperature,
            logit_bias=logit_bias or -math.log(num_classes - 1),
        )

        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.features(x)
        x = self.classifier(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.logit_layer(x)
        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)

    def get_classifier(self) -> nn.Module:
        """Returns the classifier part of the model. Note this comes before global pooling."""
        return self.classifier

    def get_feature_extractor(self) -> nn.Module:
        """Returns the feature extractor part of the model. Without global pooling."""
        return self.features


def _convnext(
    arch: str,
    block_setting: List[CNBlockConfig],
    stochastic_depth_prob: float,
    pretrained: bool,
    progress: bool,
    **kwargs: Any,
) -> BcosConvNeXt:
    model = BcosConvNeXt(
        block_setting, stochastic_depth_prob=stochastic_depth_prob, **kwargs
    )

    if pretrained:
        raise ValueError(
            "If you want to load pretrained weights, then please use the entrypoints in "
            "bcos.pretrained or bcos.model.pretrained instead."
        )

    return model


def convnext_atto(
    *, pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosConvNeXt:
    block_setting = [
        CNBlockConfig(40, 80, 2),
        CNBlockConfig(80, 160, 2),
        CNBlockConfig(160, 320, 6),
        CNBlockConfig(320, None, 2),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.1)
    return _convnext(
        "convnext_atto",
        block_setting,
        stochastic_depth_prob,
        pretrained,
        progress,
        **kwargs,
    )


def convnext_tiny(
    *, pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosConvNeXt:
    block_setting = [
        CNBlockConfig(96, 192, 3),
        CNBlockConfig(192, 384, 3),
        CNBlockConfig(384, 768, 9),
        CNBlockConfig(768, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.1)
    return _convnext(
        "convnext_tiny",
        block_setting,
        stochastic_depth_prob,
        pretrained,
        progress,
        **kwargs,
    )


def convnext_small(
    *, pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosConvNeXt:
    block_setting = [
        CNBlockConfig(96, 192, 3),
        CNBlockConfig(192, 384, 3),
        CNBlockConfig(384, 768, 27),
        CNBlockConfig(768, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.4)
    return _convnext(
        "convnext_small",
        block_setting,
        stochastic_depth_prob,
        pretrained,
        progress,
        **kwargs,
    )


def convnext_base(
    *, pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosConvNeXt:
    block_setting = [
        CNBlockConfig(128, 256, 3),
        CNBlockConfig(256, 512, 3),
        CNBlockConfig(512, 1024, 27),
        CNBlockConfig(1024, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.5)
    return _convnext(
        "convnext_base",
        block_setting,
        stochastic_depth_prob,
        pretrained,
        progress,
        **kwargs,
    )


def convnext_large(
    *, pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosConvNeXt:
    block_setting = [
        CNBlockConfig(192, 384, 3),
        CNBlockConfig(384, 768, 3),
        CNBlockConfig(768, 1536, 27),
        CNBlockConfig(1536, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.5)
    return _convnext(
        "convnext_large",
        block_setting,
        stochastic_depth_prob,
        pretrained,
        progress,
        **kwargs,
    )



import warnings
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple, Union

import torch

from ..exceptions import AmbiguityError, EMANotFound, MetricsNotFoundError
from ..structure_constants import CHECKPOINT_LAST_FILENAME
from .metric_utils import Metrics

__all__ = [
    # main
    "get_state_dict_and_training_ckpt_from_save_dir",
    "ReloadTypes",
    # minor
    "change_state_dict_keys",
    "device_safe_load_state_dict_from_path",
    # extra utils
    "get_last_checkpoint_state_dict_from_save_dir",
    "get_last_checkpoint_path_in_save_dir",
]

# type aliases
PathLike = Union[str, Path]
StateDictType = Dict[str, Any]


class ReloadTypes:
    """
    Types of reloads.
    """

    BEST = "best"
    BEST_ANY = "best_any"
    LAST = "last"
    EPOCH = "epoch_"  # important: only use this as a prefix!

    @classmethod
    def validate(cls, value: str) -> bool:
        """
        Validate the given reload type (str).
        """
        return value in [
            cls.BEST,
            cls.BEST_ANY,
            cls.LAST,
        ] or value.startswith(cls.EPOCH)


def load_training_checkpoint_for_epoch_in(
    save_dir: PathLike,
    epoch: int,
) -> StateDictType:
    """
    Load the training checkpoint for the given epoch.

    Parameters
    ----------
    save_dir : PathLike
        The directory to load the checkpoint from.
    epoch : int
        The epoch to load the checkpoint for.

    Returns
    -------
    StateDictType
        The loaded state dict.
    """
    try:
        save_path = next(Path(save_dir).glob(f"epoch={epoch}-*.ckpt"))
    except StopIteration:
        raise FileNotFoundError(
            f"Tried loading checkpoint for epoch {epoch} but none was found in {save_dir}!"
        )
    return device_safe_load_state_dict_from_path(save_path)


class PLCheckpointStateDictLoader:
    """For loading state dicts from PyTorch Lightning checkpoints."""

    MODEL_STATE_DICT_PREFIX = "model."
    MODEL_STATE_DICT_EMA_PREFIX = "ema.module."

    @classmethod
    def get_model_state_dict(
        cls,
        training_ckpt: StateDictType,
        ema: bool = False,
    ) -> StateDictType:
        model_state_dict = training_ckpt["state_dict"]
        prefix = cls.MODEL_STATE_DICT_EMA_PREFIX if ema else cls.MODEL_STATE_DICT_PREFIX

        if ema and not any(k.startswith(prefix) for k in model_state_dict.keys()):
            raise EMANotFound("EMA state dict not found in training checkpoint!")

        model_state_dict = change_state_dict_keys(
            model_state_dict, prefix_filter=prefix
        )
        return model_state_dict

    @staticmethod
    def is_suitable_state_dict(training_ckpt: StateDictType) -> bool:
        return (
            "state_dict" in training_ckpt
            and "epoch" in training_ckpt
            and "pytorch-lightning_version" in training_ckpt
        )


class SimpleCheckpointStateDictLoader:
    """For loading state dicts from simple training checkpoints (non-PL)."""

    MODEL_STATE_DICT_KEY = "model_state_dict"

    @classmethod
    def get_model_state_dict(
        cls,
        training_ckpt: StateDictType,
        **kwargs: Any,
    ) -> StateDictType:
        if kwargs:
            warnings.warn(
                f"Got unexpected kwargs: {kwargs}. Ignoring "
                f"them as they are not used in this loader."
            )
        model_state_dict = training_ckpt[cls.MODEL_STATE_DICT_KEY]
        return model_state_dict

    @classmethod
    def is_suitable_state_dict(cls, training_ckpt: StateDictType) -> bool:
        return cls.MODEL_STATE_DICT_KEY in training_ckpt


LOADERS = [
    PLCheckpointStateDictLoader,
    SimpleCheckpointStateDictLoader,
]


def load_model_state_dict_from_training_ckpt(
    training_ckpt: StateDictType,
    ema: bool = False,
) -> StateDictType:
    """
    Load the model state dict from the given training checkpoint.
    This additional indirection allows to support non-PL checkpoints,
    if required in the future.

    Parameters
    ----------
    training_ckpt : StateDictType
        The training checkpoint to load the model state dict from.
    ema : bool
        Whether to load the EMA version of the model state dict.

    Returns
    -------
    StateDictType
        The loaded model state dict.
    """
    num_suitable_loaders = sum(
        loader.is_suitable_state_dict(training_ckpt) for loader in LOADERS
    )
    if num_suitable_loaders > 1:
        raise AmbiguityError("Multiple loaders are suitable for this checkpoint!")
    elif num_suitable_loaders == 0:
        raise NotImplementedError("Unsupported checkpoint format!")

    model_state_dict = None
    for loader in LOADERS:
        if loader.is_suitable_state_dict(training_ckpt):
            model_state_dict = loader.get_model_state_dict(training_ckpt, ema=ema)
            break
    assert model_state_dict is not None, "Should have found a suitable loader!"

    return model_state_dict


def get_state_dict_and_training_ckpt_from_save_dir(
    save_dir: PathLike,
    reload: str = "last",
    ema: bool = False,
    verbose: bool = False,
) -> Tuple[StateDictType, StateDictType]:
    """
    Get the state dict of the model.

    Parameters
    ----------
    save_dir : PathLike
        The directory to load the weights/state dict from.
    reload : str
        What weights to reload. Either "last", "best", "best_any" or "epoch_<N>".
        "best_any": will check both non-ema or ema (if they exist) weights for best accuracy
        "best" will load best ema weights if `ema=True` and best non-ema otherwise
        Default: "last"
    ema : bool
        Only consider EMA weights when reloading. Not effective if `reload` is "best_any".
    verbose : bool
        Print information about loaded epoch and top-1 accuracy.
        Default: False

    Returns
    -------
    Tuple[StateDictType, StateDictType]
        A tuple of the model state dict and the training checkpoint state dict.
    """
    save_dir = Path(save_dir)
    if not save_dir.exists():
        raise FileNotFoundError(f"Directory '{save_dir}' does not exist!")
    if not save_dir.is_dir():
        raise ValueError(f"'{save_dir}' is not a directory!")
    if not ReloadTypes.validate(reload):
        raise ValueError(f"Unknown reload type: '{reload}'")

    # get the last checkpoint
    training_ckpt = get_last_checkpoint_state_dict_from_save_dir(save_dir)

    # depending on the reload get the specific checkpoint
    if reload != ReloadTypes.LAST:
        if reload == ReloadTypes.BEST or reload == ReloadTypes.BEST_ANY:
            epoch, ema = _determine_best_epoch_and_ema_status(save_dir, ema, reload)
        elif reload.startswith(ReloadTypes.EPOCH):
            # parse provided epoch
            epoch = int(reload.split("_")[1])
        else:
            raise ValueError(f"Unknown reload type: '{reload}'")

        training_ckpt = load_training_checkpoint_for_epoch_in(save_dir, epoch)

    # get the model state dict
    state_dict = load_model_state_dict_from_training_ckpt(training_ckpt, ema=ema)

    # print extra information if need be
    if verbose:
        _print_information_of_loaded_state_dict(save_dir, ema, training_ckpt)

    return state_dict, training_ckpt


def _print_information_of_loaded_state_dict(
    save_dir: PathLike, ema: bool, training_ckpt: StateDictType
) -> None:
    """
    Print information about the loaded state dict.

    Parameters
    ----------
    save_dir : PathLike
        The directory containing training checkpoints.
    ema : bool
        Whether the EMA version of the state dict was loaded.
    training_ckpt : StateDictType
        The training checkpoint from which the state dict was loaded.
    """
    epoch = training_ckpt["epoch"]
    ema_suffix = "(EMA)" if ema else ""
    print(f"Loaded epoch: {epoch} {ema_suffix}")
    try:
        metrics = Metrics.from_experiment_dir(save_dir)
        val_acc = None
        metric_key = "eval_acc1" if not ema else "eval_acc1_ema"
        # for loop in case there are multiple same numbered epochs
        for epoch_i, val_acc_i in metrics[metric_key]:
            if epoch_i == epoch:
                val_acc = val_acc_i
                break
        assert val_acc is not None, "Unable to find val acc!"
        print(f"With validation accuracy: {val_acc:.2%}")
    except MetricsNotFoundError:
        print("No validation accuracy metrics found in checkpoint!")


def _determine_best_epoch_and_ema_status(
    save_dir: PathLike,
    ema: bool,
    reload: str,
) -> Tuple[int, bool]:
    """
    Determine the best epoch and whether to use EMA weights.
    Note: EMA is only considered if `ema=True`.

    Parameters
    ----------
    save_dir : PathLike
        The save directory containing the training checkpoints.
    ema : bool
        Only consider EMA weights if `ema=True`. Not effective if `reload="best_any"`.
    reload : str
        The reload type.
        "best": will load best ema weights if `ema=True` and best non-ema otherwise
        "best_any": will check both non-ema or ema (if they exist) weights for best accuracy

    Returns
    -------
    Tuple[int, bool]
        The best epoch and whether to use EMA weights.
    """
    try:
        metrics = Metrics.from_experiment_dir(save_dir)
    except MetricsNotFoundError:
        raise MetricsNotFoundError(  # change error message
            "Unable to find metrics! These are required to find the best checkpoint!"
        )

    if reload == ReloadTypes.BEST:
        if ema:
            best_epoch, best_acc = metrics.get_best_epoch_and_accuracy_ema()
        else:
            best_epoch, best_acc = metrics.get_best_epoch_and_accuracy()
    elif reload == ReloadTypes.BEST_ANY:
        best_epoch, best_acc = metrics.get_best_epoch_and_accuracy()
        try:
            best_epoch_ema, best_acc_ema = metrics.get_best_epoch_and_accuracy_ema()
        except EMANotFound:
            best_epoch_ema, best_acc_ema = -1, -1
        if best_acc_ema > best_acc:
            best_epoch, best_acc = best_epoch_ema, best_acc_ema
            ema = True
        else:
            ema = False
    else:
        raise ValueError(
            f"Unknown reload type: '{reload}'. Only 'best' or 'best_any' supported!"
        )

    return best_epoch, ema


def change_state_dict_keys(
    state_dict: StateDictType,
    prefix_filter: Optional[str] = None,
    filter_fn: Optional[Callable[[str], bool]] = None,
    remove_prefix: bool = True,
) -> StateDictType:
    """
    Change the keys of a state dict.

    Parameters
    ----------
    state_dict : Dict[str, Any]
        The dict containing the state.
    prefix_filter: Optional[str]
        Optional prefix to filter the state dict by. Default: None
    filter_fn : Optional[Callable[[str], bool]]
        Optional function to filter the state dict by. Default: None
    remove_prefix : bool
        Whether to remove the prefix from the state dict's keys.
        Only works if `prefix_filter` is set. Default: True.

    Returns
    -------
    StateDictType
        The state dict with the changed keys.
    """
    # make sure both filters are not set
    if prefix_filter is not None and filter_fn is not None:
        raise ValueError("Only one of `prefix_filter` and `filter_fn` can be set.")

    # create filter function
    def _filter_fn(key: str) -> bool:
        if prefix_filter is not None:
            return key.startswith(prefix_filter)
        elif filter_fn is not None:
            return filter_fn(key)
        else:
            return True

    # optionally remove prefix
    def _key_mutator(key: str) -> str:
        if remove_prefix and prefix_filter is not None:
            return key[len(prefix_filter) :]
        else:
            return key

    # filter state dict
    state_dict = {
        _key_mutator(key): value for key, value in state_dict.items() if _filter_fn(key)
    }

    if len(state_dict) == 0 and (prefix_filter is not None or filter_fn is not None):
        warnings.warn("Filtering possibly resulted in an empty model state dict!")
    return state_dict


def device_safe_load_state_dict_from_path(path: PathLike) -> StateDictType:
    """
    Load state dict from path, but make sure it's on the CPU.

    Parameters
    ----------
    path : PathLike
        The path to load the state dict from.

    Returns
    -------
    StateDictType
        The loaded state dict.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(
            f"Tried loading last checkpoint but none found at: '{path}'"
        )

    state_dict = torch.load(path, map_location="cpu")

    return state_dict


def get_last_checkpoint_path_in_save_dir(save_dir: PathLike) -> Path:
    """
    Get the path of the last checkpoint in the given directory.
    """
    return Path(save_dir) / CHECKPOINT_LAST_FILENAME


def get_last_checkpoint_state_dict_from_save_dir(save_dir: PathLike) -> StateDictType:
    """
    Try to load the last checkpoint state dict.
    """
    save_path = get_last_checkpoint_path_in_save_dir(save_dir)
    if not save_path.exists():
        raise FileNotFoundError(
            f"Tried loading last checkpoint but none found at: '{save_path}'"
        )
    return device_safe_load_state_dict_from_path(save_path)



import math  # noqa

from bcos.data.presets import (
    ImageNetClassificationPresetEval,
    ImageNetClassificationPresetTrain,
)
from bcos.experiments.utils import (
    configs_cli,
    create_configs_with_different_seeds,
    update_config,
)
from bcos.modules import norms
from bcos.modules.losses import (
    BinaryCrossEntropyLoss,
    UniformOffLabelsBCEWithLogitsLoss,
)
from bcos.optim import LRSchedulerFactory, OptimizerFactory

__all__ = ["CONFIGS"]

NUM_CLASSES = 1_000

# These are mainly based on the recipes from
# https://github.com/pytorch/vision/blob/93723b481d1f6e/references/classification/README.md
DEFAULT_BATCH_SIZE = 64  # per GPU! * 4 = 256 effective
DEFAULT_NUM_EPOCHS = 90
DEFAULT_LR = 1e-3
DEFAULT_CROP_SIZE = 224

DEFAULT_NORM_LAYER = norms.NoBias(norms.BatchNormUncentered2d)  # bnu-linear

DEFAULT_OPTIMIZER = OptimizerFactory(name="Adam", lr=DEFAULT_LR)
DEFAULT_LR_SCHEDULE = LRSchedulerFactory(
    name="cosineannealinglr",
    epochs=DEFAULT_NUM_EPOCHS,
    warmup_method="linear",
    warmup_epochs=5,
    warmup_decay=0.01,
)

DEFAULTS = dict(
    data=dict(
        train_transform=ImageNetClassificationPresetTrain(
            crop_size=DEFAULT_CROP_SIZE,
            is_bcos=True,
        ),
        test_transform=ImageNetClassificationPresetEval(
            crop_size=DEFAULT_CROP_SIZE,
            is_bcos=True,
        ),
        batch_size=DEFAULT_BATCH_SIZE,
        num_workers=16,
        num_classes=NUM_CLASSES,
    ),
    model=dict(
        is_bcos=True,
        # "name": None,
        args=dict(
            num_classes=NUM_CLASSES,
            norm_layer=DEFAULT_NORM_LAYER,
            logit_bias=-math.log(NUM_CLASSES - 1),
        ),
        bcos_args=dict(
            b=2,
            max_out=1,
        ),
    ),
    criterion=UniformOffLabelsBCEWithLogitsLoss(),
    test_criterion=BinaryCrossEntropyLoss(),
    optimizer=DEFAULT_OPTIMIZER,
    lr_scheduler=DEFAULT_LR_SCHEDULE,
    trainer=dict(
        max_epochs=DEFAULT_NUM_EPOCHS,
    ),
    use_agc=True,
)


# helper
def update_default(new_config):
    return update_config(DEFAULTS, new_config)


RESNET_DEPTHS = [18, 34, 50, 101, 152]
resnets = {
    f"resnet_{depth}": update_default(
        dict(
            model=dict(
                name=f"resnet{depth}",
            ),
        )
    )
    for depth in RESNET_DEPTHS
}

RESNEXT_DEPTHS_AND_WIDTHS = ["50_32x4d"]
resnexts = {
    f"resnext_{depth_and_width}": update_default(
        dict(
            model=dict(
                name=f"resnext{depth_and_width}",
            ),
            lr_scheduler=DEFAULT_LR_SCHEDULE.with_epochs(100),
            trainer=dict(
                max_epochs=100,
            ),
        )
    )
    for depth_and_width in RESNEXT_DEPTHS_AND_WIDTHS
}

DENSENET_DEPTHS = [121, 161, 201, 169]
densenets = {
    f"densenet_{depth}": update_default(
        dict(
            model=dict(
                name=f"densenet{depth}",
            ),
        )
    )
    for depth in DENSENET_DEPTHS
}

VGG_DEPTHS = [11]
vggs_bnu = {
    f"vgg_{depth}_bnu": update_default(
        dict(
            model=dict(
                name=f"vgg{depth}_bnu",
            )
        )
    )
    for depth in VGG_DEPTHS
}

# -------------------------------------------------------------------------

CONFIGS = dict()
CONFIGS.update(resnets)
CONFIGS.update(resnexts)
CONFIGS.update(densenets)
CONFIGS.update(vggs_bnu)
CONFIGS.update(create_configs_with_different_seeds(CONFIGS, seeds=[420, 1337]))

if __name__ == "__main__":
    configs_cli(CONFIGS)



"""
Position Normalization Uncentered (PNU) for 2D inputs

Positional Normalization:
https://github.com/Boyiliee/Positional-Normalization
"""
import torch
from torch import Tensor, nn

from bcos.modules.common import DetachableModule

__all__ = [
    "PositionNormUncentered2d",
]


class PositionNormUncentered2d(nn.LayerNorm, DetachableModule):
    def __init__(
        self,
        features: int,
        eps: float = 1e-5,
        affine: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        assert isinstance(
            features, int
        ), f"Provide #features as an int not {type(features)=}"
        DetachableModule.__init__(self)
        super().__init__(
            normalized_shape=features,
            eps=eps,
            elementwise_affine=affine,
            device=device,
            dtype=dtype,
        )
        self.features = features

    def forward(self, x: Tensor) -> Tensor:
        assert x.dim() == 4, f"input should be 4D not {x.dim()}D"

        # ------------ manual PNU forward pass -------------

        # get stats
        var = torch.var(x, dim=1, unbiased=False, keepdim=True)
        if self.detach:
            var = var.detach()
        std = (var + self.eps).sqrt()

        x = x / std

        # affine transformation
        if self.weight is not None:
            x = self.weight[None, ..., None, None] * x
        if self.bias is not None:
            x = x + self.bias[None, ..., None, None]

        return x



"""
B-cos DenseNet models

Modified from https://github.com/pytorch/vision/blob/0504df5ddf9431909130e7788faf05446bb8/torchvision/models/densenet.py

CIFAR10 modifications from
https://github.com/gpleiss/efficient_densenet_pytorch/blob/master/models/densenet.py
"""
import math
import re
from collections import OrderedDict
from typing import Any, Callable, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as cp
from torch import Tensor
from torch.hub import load_state_dict_from_url

from bcos.common import BcosUtilMixin
from bcos.modules import BcosConv2d, LogitLayer, norms

__all__ = [
    "BcosDenseNet",
    "densenet121",
    "densenet169",
    "densenet201",
    "densenet161",
    # c10
    "cifar10_densenet250",
    "cifar10_densenet100",
    "cifar10_densenet190",
]


DEFAULT_NORM_LAYER = norms.NoBias(norms.DetachablePositionNorm2d)
DEFAULT_CONV_LAYER = BcosConv2d


class _DenseLayer(nn.Module):
    def __init__(
        self,
        num_input_features: int,
        growth_rate: int,
        bn_size: int,
        drop_rate: float,
        memory_efficient: bool = False,
        norm_layer: Callable[..., nn.Module] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
    ) -> None:
        super(_DenseLayer, self).__init__()
        # Diff to torchvision: Removed ReLU and replaced BatchNorm with norm_layer
        self.norm1 = norm_layer(num_input_features)
        # Diff to torchvision: Replace Conv2d with ProjectionConv2d
        self.conv1 = conv_layer(
            num_input_features,
            bn_size * growth_rate,
            kernel_size=1,
            stride=1,
            padding=0,
        )
        self.norm2 = norm_layer(bn_size * growth_rate)
        self.conv2 = conv_layer(
            bn_size * growth_rate,
            growth_rate,
            kernel_size=3,
            stride=1,
            padding=1,
        )
        # Diff End
        # Diff End
        self.drop_rate = float(drop_rate)
        self.memory_efficient = memory_efficient

    def bn_function(self, inputs: List[Tensor]) -> Tensor:
        concated_features = torch.cat(inputs, 1)
        # Diff to torchvision: Deleted relu
        bottleneck_output = self.conv1(self.norm1(concated_features))  # noqa: T484
        # Diff End
        return bottleneck_output

    # todo: rewrite when torchscript supports any
    def any_requires_grad(self, input: List[Tensor]) -> bool:
        for tensor in input:
            if tensor.requires_grad:
                return True
        return False

    @torch.jit.unused  # noqa: T484
    def call_checkpoint_bottleneck(self, input: List[Tensor]) -> Tensor:
        def closure(*inputs):
            return self.bn_function(inputs)

        return cp.checkpoint(closure, *input)

    @torch.jit._overload_method  # noqa: F811
    def forward(self, input: List[Tensor]) -> Tensor:
        pass

    @torch.jit._overload_method  # noqa: F811
    def forward(self, input: Tensor) -> Tensor:  # noqa: F811
        pass

    # torchscript does not yet support *args, so we overload method
    # allowing it to take either a List[Tensor] or single Tensor
    def forward(self, input: Tensor) -> Tensor:  # noqa: F811
        if isinstance(input, Tensor):
            prev_features = [input]
        else:
            prev_features = input

        if self.memory_efficient and self.any_requires_grad(prev_features):
            if torch.jit.is_scripting():
                raise Exception("Memory Efficient not supported in JIT")

            bottleneck_output = self.call_checkpoint_bottleneck(prev_features)
        else:
            bottleneck_output = self.bn_function(prev_features)
        # Diff to torchvision: Deleted relu
        new_features = self.conv2(self.norm2(bottleneck_output))
        # Diff End

        if self.drop_rate > 0:
            new_features = F.dropout(
                new_features, p=self.drop_rate, training=self.training
            )
        return new_features


class _DenseBlock(nn.ModuleDict):
    _version = 2

    def __init__(
        self,
        num_layers: int,
        num_input_features: int,
        bn_size: int,
        growth_rate: int,
        drop_rate: float,
        memory_efficient: bool = False,
        norm_layer: Callable[..., nn.Module] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
    ) -> None:
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(
                num_input_features + i * growth_rate,
                growth_rate=growth_rate,
                bn_size=bn_size,
                drop_rate=drop_rate,
                memory_efficient=memory_efficient,
                norm_layer=norm_layer,
                conv_layer=conv_layer,
            )
            self.add_module("denselayer%d" % (i + 1), layer)

    def forward(self, init_features: Tensor) -> Tensor:
        features = [init_features]
        for name, layer in self.items():
            new_features = layer(features)
            features.append(new_features)
        return torch.cat(features, 1)


class _Transition(nn.Sequential):
    def __init__(
        self,
        num_input_features: int,
        num_output_features: int,
        norm_layer: Callable[..., nn.Module] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
    ) -> None:
        super(_Transition, self).__init__()
        # Diff to torchvision: Deleted relu and changed conv to ProjectionConv2d
        self.add_module("norm", norm_layer(num_input_features))
        self.conv = conv_layer(
            num_input_features,
            num_output_features,
            kernel_size=1,
            stride=1,
            padding=0,
        )
        # Diff End

        self.add_module("pool", nn.AvgPool2d(kernel_size=2, stride=2))


class BcosDenseNet(BcosUtilMixin, nn.Module):
    r"""Densenet-BC model class, based on
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_.

    Args:
        growth_rate (int) - how many filters to add each layer (`k` in paper)
        block_config (list of 4 ints) - how many layers in each pooling block
        num_init_features (int) - the number of filters to learn in the first convolution layer
        bn_size (int) - multiplicative factor for number of bottle neck layers
          (i.e. bn_size * k features in the bottleneck layer)
        drop_rate (float) - dropout rate after each dense layer
        num_classes (int) - number of classification classes
        memory_efficient (bool) - If True, uses checkpointing. Much more memory efficient,
          but slower. Default: *False*. See `"paper" <https://arxiv.org/pdf/1707.06990.pdf>`_.
    """

    def __init__(
        self,
        growth_rate: int = 32,
        block_config: Tuple[int, int, int, int] = (6, 12, 24, 16),
        num_init_features: int = 64,
        bn_size: int = 4,
        drop_rate: float = 0,
        num_classes: int = 1000,
        in_chans: int = 6,
        memory_efficient: bool = False,
        norm_layer: Callable[..., nn.Module] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
        small_inputs: bool = False,  # True for 32x32 images from gpleiss' impl
        logit_bias: Optional[float] = None,
        logit_temperature: Optional[float] = None,
    ) -> None:
        super(BcosDenseNet, self).__init__()

        # First convolution
        # Diff to torchvision: Deleted ReLU, changed Conv2d for ProjectionConv2d and
        # MaxPool for AvgPool
        if small_inputs:
            self.features = nn.Sequential(
                OrderedDict(
                    [
                        (
                            "conv0",
                            conv_layer(
                                in_chans,
                                num_init_features,
                                kernel_size=3,
                                stride=1,
                                padding=1,
                            ),
                        )
                    ]
                )
            )
        else:
            self.features = nn.Sequential(
                OrderedDict(
                    [
                        (
                            "conv0",
                            conv_layer(
                                in_chans,
                                num_init_features,
                                kernel_size=7,
                                stride=2,
                                padding=3,
                            ),
                        ),
                        ("norm0", norm_layer(num_init_features)),
                        ("pool0", nn.AvgPool2d(kernel_size=3, stride=2, padding=1)),
                    ]
                )
            )
        # Diff End
        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=bn_size,
                growth_rate=growth_rate,
                drop_rate=drop_rate,
                memory_efficient=memory_efficient,
                norm_layer=norm_layer,
                conv_layer=conv_layer,
            )
            self.features.add_module("denseblock%d" % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(
                    num_input_features=num_features,
                    num_output_features=num_features // 2,
                    norm_layer=norm_layer,
                    conv_layer=conv_layer,
                )
                self.features.add_module("transition%d" % (i + 1), trans)
                num_features = num_features // 2

        # Final norm layer
        self.features.add_module("norm5", norm_layer(num_features))

        # Diff to torchvision: changed Linear layer to BcosConv (conv classifier)
        self.classifier = conv_layer(num_features, num_classes, kernel_size=1)
        self.num_classes = num_classes
        self.logit_layer = LogitLayer(
            logit_temperature=logit_temperature,
            logit_bias=logit_bias or -math.log(num_classes - 1),
        )
        # Diff End

        # Official init from torch repo.
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, x: Tensor) -> Tensor:
        # Diff to torchvision: Deleted relu
        features = self.features(x)
        # out = F.relu(features, inplace=True)
        out = self.classifier(features)
        out = F.adaptive_avg_pool2d(out, (1, 1))
        out = torch.flatten(out, 1)
        out = self.logit_layer(out)
        # Diff End
        return out

    def get_classifier(self) -> nn.Module:
        """Returns the classifier part of the model. Note this comes before global pooling."""
        return self.classifier

    def get_feature_extractor(self) -> nn.Module:
        """Returns the feature extractor part of the model. Without global pooling."""
        return self.features


def _load_state_dict(model: nn.Module, model_url: str, progress: bool) -> None:
    # '.'s are no longer allowed in module names, but previous _DenseLayer
    # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
    # They are also in the checkpoints in model_urls. This pattern is used
    # to find such keys.
    pattern = re.compile(
        r"^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$"
    )

    state_dict = load_state_dict_from_url(
        model_url, map_location="cpu", progress=progress, check_hash=True
    )
    for key in list(state_dict.keys()):
        res = pattern.match(key)
        if res:
            new_key = res.group(1) + res.group(2)
            state_dict[new_key] = state_dict[key]
            del state_dict[key]
    model.load_state_dict(state_dict)


def _densenet(
    arch: str,
    growth_rate: int,
    block_config: Union[Tuple[int, int, int], Tuple[int, int, int, int]],
    num_init_features: int,
    pretrained: bool,
    progress: bool,
    **kwargs: Any,
) -> BcosDenseNet:
    model = BcosDenseNet(growth_rate, block_config, num_init_features, **kwargs)
    if pretrained:
        raise ValueError(
            "If you want to load pretrained weights, then please use the entrypoints in "
            "bcos.pretrained or bcos.model.pretrained instead."
        )
    return model


def densenet121(
    pretrained: bool = False,
    progress: bool = True,
    num_init_features=64,
    growth_rate=32,
    **kwargs: Any,
) -> BcosDenseNet:
    return _densenet(
        "densenet121",
        growth_rate,
        (6, 12, 24, 16),
        num_init_features,
        pretrained,
        progress,
        **kwargs,
    )


def densenet161(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    return _densenet(
        "densenet161", 48, (6, 12, 36, 24), 96, pretrained, progress, **kwargs
    )


def densenet169(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    return _densenet(
        "densenet169", 32, (6, 12, 32, 32), 64, pretrained, progress, **kwargs
    )


def densenet201(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    return _densenet(
        "densenet201", 32, (6, 12, 48, 32), 64, pretrained, progress, **kwargs
    )


# cifar helpers
def _get_cifar_block_config(depth: int) -> Tuple[int, int, int]:
    if (depth - 4) % 3:
        raise Exception("Invalid depth!")
    return ((depth - 4) // 6,) * 3


def _update_default_cifar(kwargs: dict) -> None:
    kwargs.setdefault("num_classes", 10)
    kwargs.setdefault("small_inputs", True)


def cifar10_densenet52(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    L = 52
    k = 12
    block_config = _get_cifar_block_config(L)
    _update_default_cifar(kwargs)
    return _densenet(
        f"cifar10_densenet{L}", k, block_config, k * 2, pretrained, progress, **kwargs
    )


# cifar densenet configs from Sec. 3#Implementation detail from the paper
def cifar10_densenet100(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    L = 100
    k = 12
    block_config = _get_cifar_block_config(L)
    _update_default_cifar(kwargs)
    return _densenet(
        f"cifar10_densenet{L}", k, block_config, k * 2, pretrained, progress, **kwargs
    )


def cifar10_densenet250(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    L = 250
    k = 24
    block_config = _get_cifar_block_config(L)
    _update_default_cifar(kwargs)
    return _densenet(
        f"cifar10_densenet{L}", k, block_config, k * 2, pretrained, progress, **kwargs
    )


def cifar10_densenet190(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosDenseNet:
    L = 190
    k = 40
    block_config = _get_cifar_block_config(L)
    _update_default_cifar(kwargs)
    return _densenet(
        f"cifar10_densenet{L}", k, block_config, k * 2, pretrained, progress, **kwargs
    )



"""
B-cos ResNet/ResNeXt/Wide-ResNet models

Modified from https://github.com/pytorch/vision/blob/0504df5ddf9431909130e7788faf05446bb8a2/torchvision/models/resnet.py

CIFAR10 modifications from
https://github.com/chenyaofo/pytorch-cifar-models/blob/e9482ebc665084761ad9c84d36c83cbb82/pytorch_cifar_models/resnet.py
"""
import math
from typing import Any, Callable, List, Optional, Type, Union

import torch.nn as nn
from torchvision.ops import StochasticDepth

from bcos.common import BcosUtilMixin
from bcos.modules import BcosConv2d, LogitLayer, norms

__all__ = [
    "BcosResNet",
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "resnext50_32x4d",
    "resnext101_32x8d",
    "wide_resnet50_2",
    "wide_resnet101_2",
    # cifar
    "cifar10_resnet20",
    "cifar10_resnet32",
    "cifar10_resnet44",
    "cifar10_resnet56",
    "cifar10_resnet110",
    "cifar10_resnext29_8x64d",
    "cifar10_resnext29_16x64d",
    "cifar10_resnext29_32x4d",
    "cifar10_resnext29_16x8d",
    "cifar10_resnext47_16x8d",
    "cifar10_resnext47_32x4d",
    "cifar10_resnext65_16x8d",
    "cifar10_resnext65_32x4d",
    "cifar10_resnext101_16x8d",
    "cifar10_resnext101_32x4d",
]


DEFAULT_NORM_LAYER = norms.NoBias(norms.DetachablePositionNorm2d)
DEFAULT_CONV_LAYER = BcosConv2d


def conv3x3(
    in_planes: int,
    out_planes: int,
    stride: int = 1,
    groups: int = 1,
    dilation: int = 1,
    conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
):
    """3x3 convolution with padding"""
    return conv_layer(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(
    in_planes: int,
    out_planes: int,
    stride: int = 1,
    conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
):
    """1x1 convolution"""
    return conv_layer(
        in_planes,
        out_planes,
        kernel_size=1,
        stride=stride,
        bias=False,
    )


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
        # act_layer: Callable[..., nn.Module] = None,
        stochastic_depth_prob: float = 0.0,
    ):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.Identity
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(
            inplanes,
            planes,
            stride,
            conv_layer=conv_layer,
        )
        self.bn1 = norm_layer(planes)
        # self.act = act_layer(inplace=True)
        self.conv2 = conv3x3(
            planes,
            planes,
            conv_layer=conv_layer,
        )
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride
        self.stochastic_depth = (
            StochasticDepth(stochastic_depth_prob, "row")
            if stochastic_depth_prob
            else None
        )

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        # out = self.act(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.stochastic_depth is not None:
            out = self.stochastic_depth(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        # out = self.act(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
        # act_layer: Callable[..., nn.Module] = None,
        stochastic_depth_prob: float = 0.0,
    ) -> None:
        super(Bottleneck, self).__init__()
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(
            inplanes,
            width,
            conv_layer=conv_layer,
        )
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(
            width,
            width,
            stride,
            groups,
            dilation,
            conv_layer=conv_layer,
        )
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(
            width,
            planes * self.expansion,
            conv_layer=conv_layer,
        )
        self.bn3 = norm_layer(planes * self.expansion)
        # self.act = act_layer(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.stochastic_depth = (
            StochasticDepth(stochastic_depth_prob, "row")
            if stochastic_depth_prob
            else None
        )

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        # out = self.act(out)

        out = self.conv2(out)
        out = self.bn2(out)
        # out = self.act(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.stochastic_depth is not None:
            out = self.stochastic_depth(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        # out = self.act(out)

        return out


class BcosResNet(BcosUtilMixin, nn.Module):
    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        num_classes: int = 1000,
        in_chans: int = 6,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = DEFAULT_NORM_LAYER,
        conv_layer: Callable[..., nn.Module] = DEFAULT_CONV_LAYER,
        # act_layer: Callable[..., nn.Module] = None,
        inplanes: int = 64,
        small_inputs: bool = False,
        stochastic_depth_prob: float = 0.0,
        logit_bias: Optional[float] = None,
        logit_temperature: Optional[float] = None,
        **kwargs: Any,  # ignore rest
    ):
        super().__init__()

        if kwargs:
            print("The following args passed to model will be ignored", kwargs)

        if norm_layer is None:
            norm_layer = nn.Identity
        self._norm_layer = norm_layer
        self._conv_layer = conv_layer
        # self._act_layer = act_layer

        self.inplanes = inplanes
        self.dilation = 1
        n = len(layers)  # number of stages
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False] * (n - 1)
        if len(replace_stride_with_dilation) != n - 1:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                f"or a {n - 1}-element tuple, got {replace_stride_with_dilation}"
            )
        self.groups = groups
        self.base_width = width_per_group

        if small_inputs:
            self.conv1 = conv3x3(
                in_chans,
                self.inplanes,
                conv_layer=conv_layer,
            )
            self.pool = None
        else:
            self.conv1 = conv_layer(
                in_chans,
                self.inplanes,
                kernel_size=7,
                stride=2,
                padding=3,
            )
            self.pool = nn.AvgPool2d(kernel_size=3, stride=2, padding=1)

        self.bn1 = norm_layer(self.inplanes)
        # self.act = act_layer(inplace=True)

        self.__total_num_blocks = sum(layers)
        self.__num_blocks = 0
        self.layer1 = self._make_layer(
            block,
            inplanes,
            layers[0],
            stochastic_depth_prob=stochastic_depth_prob,
        )
        self.layer2 = self._make_layer(
            block,
            inplanes * 2,
            layers[1],
            stride=2,
            dilate=replace_stride_with_dilation[0],
            stochastic_depth_prob=stochastic_depth_prob,
        )
        self.layer3 = self._make_layer(
            block,
            inplanes * 4,
            layers[2],
            stride=2,
            dilate=replace_stride_with_dilation[1],
            stochastic_depth_prob=stochastic_depth_prob,
        )
        try:
            self.layer4 = self._make_layer(
                block,
                inplanes * 8,
                layers[3],
                stride=2,
                dilate=replace_stride_with_dilation[2],
                stochastic_depth_prob=stochastic_depth_prob,
            )
            last_ch = inplanes * 8
        except IndexError:
            self.layer4 = None
            last_ch = inplanes * 4

        self.num_features = last_ch * block.expansion
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.num_classes = num_classes
        self.fc = conv_layer(
            self.num_features,
            self.num_classes,
            kernel_size=1,
        )
        self.logit_layer = LogitLayer(
            logit_temperature=logit_temperature,
            logit_bias=logit_bias or -math.log(num_classes - 1),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck) and m.bn3.weight is not None:
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock) and m.bn2.weight is not None:
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
        stochastic_depth_prob: float = 0.0,
    ) -> nn.Sequential:
        norm_layer = self._norm_layer
        conv_layer = self._conv_layer
        # act_layer = self._act_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(
                    self.inplanes,
                    planes * block.expansion,
                    stride,
                    conv_layer=conv_layer,
                ),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer=norm_layer,
                conv_layer=conv_layer,
                # act_layer=act_layer,
                stochastic_depth_prob=stochastic_depth_prob
                * self.__num_blocks
                / (self.__total_num_blocks - 1),
            )
        )
        self.__num_blocks += 1
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                    conv_layer=conv_layer,
                    # act_layer=act_layer,
                    stochastic_depth_prob=stochastic_depth_prob
                    * self.__num_blocks
                    / (self.__total_num_blocks - 1),
                )
            )
            self.__num_blocks += 1

        return nn.Sequential(*layers)

    def forward_features(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        # x = self.act(x)
        if self.pool is not None:
            x = self.pool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        if self.layer4 is not None:
            x = self.layer4(x)
        return x

    def forward(self, x):
        x = self.forward_features(x)

        x = self.fc(x)
        x = self.avgpool(x)
        x = x.flatten(1)
        x = self.logit_layer(x)

        return x

    def get_classifier(self) -> nn.Module:
        """Returns the classifier part of the model. Note this comes before global pooling."""
        return self.fc

    def get_feature_extractor(self) -> nn.Module:
        """Returns the feature extractor part of the model. Without global pooling."""
        modules = [
            self.conv1,
            self.bn1,
            # self.act,
        ]
        if self.pool is not None:
            modules += [self.pool]
        modules += [
            self.layer1,
            self.layer2,
            self.layer3,
        ]
        if self.layer4 is not None:
            modules += [self.layer4]
        return nn.Sequential(*modules)


def _resnet(
    arch: str,
    block: Type[Union[BasicBlock, Bottleneck]],
    layers: List[int],
    pretrained: bool = False,
    progress: bool = True,
    inplanes: int = 64,
    **kwargs: Any,
) -> BcosResNet:
    model = BcosResNet(block, layers, inplanes=inplanes, **kwargs)
    if pretrained:
        raise ValueError(
            "If you want to load pretrained weights, then please use the entrypoints in "
            "bcos.pretrained or bcos.model.pretrained instead."
        )
    return model


# ---------------------
# ResNets for ImageNet
# ---------------------
def resnet18(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    return _resnet("resnet18", BasicBlock, [2, 2, 2, 2], pretrained, progress, **kwargs)


def resnet34(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    return _resnet("resnet34", BasicBlock, [3, 4, 6, 3], pretrained, progress, **kwargs)


def resnet50(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    return _resnet("resnet50", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs)


def resnet101(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    return _resnet(
        "resnet101", Bottleneck, [3, 4, 23, 3], pretrained, progress, **kwargs
    )


def resnet152(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    return _resnet(
        "resnet152", Bottleneck, [3, 8, 36, 3], pretrained, progress, **kwargs
    )


def resnext50_32x4d(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    return _resnet(
        "resnext50_32x4d", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def resnext101_32x8d(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 8
    return _resnet(
        "resnext101_32x8d",
        Bottleneck,
        [3, 4, 23, 3],
        pretrained,
        progress,
        **kwargs,
    )


def wide_resnet50_2(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    kwargs["width_per_group"] = 64 * 2
    return _resnet(
        "wide_resnet50_2", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def wide_resnet101_2(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> BcosResNet:
    kwargs["width_per_group"] = 64 * 2
    return _resnet(
        "wide_resnet101_2",
        Bottleneck,
        [3, 4, 23, 3],
        pretrained,
        progress,
        **kwargs,
    )


# ---------------------
# ResNets for CIFAR-10
# ---------------------
def _update_if_not_present(key, value, d):
    if key not in d:
        d[key] = value


def _update_default_cifar(kwargs) -> None:
    _update_if_not_present("num_classes", 10, kwargs)
    _update_if_not_present("small_inputs", True, kwargs)


def cifar10_resnet20(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnet20",
        BasicBlock,
        [3] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnet32(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnet32",
        BasicBlock,
        [5] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnet44(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnet44",
        BasicBlock,
        [7] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnet56(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnet56",
        BasicBlock,
        [9] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnet110(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnet110",
        BasicBlock,
        [18] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


# --------------------
# ResNeXts
# --------------------
# These are model configs specified in the ResNeXt Paper https://arxiv.org/pdf/1611.05431.pdf


def cifar10_resnext29_8x64d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 8
    kwargs["width_per_group"] = 64
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext29_8x64d",
        Bottleneck,
        [3] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=64,
        **kwargs,
    )


def cifar10_resnext29_16x64d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 16
    kwargs["width_per_group"] = 64
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext29_16x64d",
        Bottleneck,
        [3] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=64,
        **kwargs,
    )


# The model configs specified in the ResNeXt Paper
# are very large. We use smaller ones here.
# So instead of the 8x64d or 16x64d settings (with [64, 128, 256] widths)
# we have 32x4d and 16x8d settings ([16, 32, 64]).
# first 32x4d


def cifar10_resnext29_32x4d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext29_32x4d",
        Bottleneck,
        [3] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnext47_32x4d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext47_32x4d",
        Bottleneck,
        [5] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnext65_32x4d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext65_32x4d",
        Bottleneck,
        [7] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnext101_32x4d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext101_32x4d",
        Bottleneck,
        [11] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


# next with 16x8d


def cifar10_resnext29_16x8d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 16
    kwargs["width_per_group"] = 8
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext29_32x4d",
        Bottleneck,
        [3] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnext47_16x8d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 16
    kwargs["width_per_group"] = 8
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext47_32x4d",
        Bottleneck,
        [5] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnext65_16x8d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 16
    kwargs["width_per_group"] = 8
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext65_32x4d",
        Bottleneck,
        [7] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )


def cifar10_resnext101_16x8d(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> BcosResNet:
    kwargs["groups"] = 16
    kwargs["width_per_group"] = 8
    _update_default_cifar(kwargs)
    return _resnet(
        "cifar10_resnext101_32x4d",
        Bottleneck,
        [11] * 3,
        pretrained=pretrained,
        progress=progress,
        inplanes=16,
        **kwargs,
    )



"""
General settings. Mainly paths to data.
"""
import os

# data root (mainly for CIFAR10)
DATA_ROOT = os.getenv("DATA_ROOT")

# ImageNet path
IMAGENET_PATH = os.getenv("IMAGENET_PATH")

# ---------------------------------------------
# Following are only needed for caching!!!
# ---------------------------------------------
SHMTMPDIR = "/dev/shm"

# ImageNet train data tar files
# Note this is not the same as the official ImageNet tar file
# this expects each class to inside its own tar file, which will then be extracted
# to SHMTMPDIR
IMAGENET_TRAIN_TAR_FILES_DIR = os.getenv("IMAGENET_TRAIN_TAR_FILES_DIR")

# Path to redis server
REDIS_SERVER = os.getenv("REDIS_SERVER_PATH")



"""
This module contains utilities related to B-cos models.
None of this is "essential" to training or doing inference with the models.
(Most of the stuff can be done quickly and easily in a few lines of code.)
However, they are useful for e.g. visualizing the explanations etc.
So essentially it's a collection of convenience/helper functions/classes.
"""
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    # this isn't supposed to be a hard dependency
    import matplotlib
    import matplotlib.pyplot as plt

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

if torch.__version__ < "2.0":
    from torch.autograd.grad_mode import _DecoratorContextManager  # noqa
else:
    from torch.utils._contextlib import _DecoratorContextManager  # noqa

__all__ = [
    "BcosUtilMixin",
    "explanation_mode",
    "gradient_to_image",
    "plot_contribution_map",
]


TensorLike = Union[Tensor, np.ndarray]


class BcosUtilMixin:
    """
    This mixin defines useful helpers for dealing with explanations.
    This is just a convenience to attach useful B-cos specific functionality.

    The parameters to ``__init__`` are just passed to the actual base class (e.g. ``torch.nn.Module``).

    Notes
    -----
    Since this is a mixin, if you want to use this, you need to inherit from this first
    and then from the actual base class (e.g. `torch.nn.Module`).


    Examples
    --------
    >>> from bcos.modules import BcosConv2d
    >>> class MyModel(BcosUtilMixin, torch.nn.Module):
    ...     def __init__(self, in_chan: int, out_chan: int):
    ...         super().__init__()
    ...         self.linear = BcosConv2d(in_chan, out_chan, 3)
    ...     def forward(self, x: torch.Tensor) -> torch.Tensor:
    ...         return self.linear(x)
    >>> model = MyModel(6, 16)
    >>> model.explain(torch.rand(1, 6, 32, 32))  # get explain method
    >>> with model.explanation_mode():  # explanation mode ctx (assuming we have detachable modules)
    ...     ...  # do something with explanation mode activated

    Parameters
    ----------
    args: Any
        Positional arguments to pass to the parent class.
    kwargs: Any
        Keyword arguments to pass to the parent class.
    """

    to_probabilities = torch.sigmoid
    """ Function to convert model outputs to probabilties. """

    def __init__(self, *args: Any, **kwargs: Any):
        self.__explanation_mode_ctx = explanation_mode(self)  # type: ignore
        super().__init__(*args, **kwargs)

    def explanation_mode(self) -> "explanation_mode":
        """
        Returns a context manager which puts model in explanation mode
        and when exiting puts it in normal mode back again.

        Returns
        -------
        explanation_mode
            The context manager which puts model in and out to/from explanation mode.
        """
        return self.__explanation_mode_ctx

    def explain(
        self,
        in_tensor,
        idx=None,
        **grad2img_kwargs,
    ) -> "Dict[str, Any]":
        """
        Generates an explanation for the given input tensor.
        This is not a generic explanation method, but rather a helper for simply getting explanations.
        It is intended for simple use cases (simple exploration, debugging, etc.).

        Parameters
        ----------
        in_tensor : Tensor
            The input tensor to explain. Must be 4-dimensional and have batch size of 1.
        idx : int, optional
            The index of the output to explain. If None, the prediction is explained.
        grad2img_kwargs : Any
            Additional keyword arguments passed to `gradient_to_image` method
            for generating the explanation.

        Examples
        --------
        Here is an example of how to use this method to generate and visualize an explanation and a contribution map:

        >>> model = ...  # instantiate some B-cos model
        >>> img = ...  # instantiate some input image tensor
        >>> expl_out = model.explain(img)
        >>> expl_out["prediction"]
        932

        >>> import matplotlib.pyplot as plt
        >>> plt.imshow(expl_out["explanation"])
        >>> plt.show()  # show the explanation

        >>> model.plot_contribution_map(expl_out["contribution_map"])
        >>> plt.show()  # show the contribution map

        Warnings
        --------
        This method is NOT optimized for speed.
        Also, on a more general note: Care should be taken when generating explanations during training,
        as the gradients might be different from during inference.

        Returns
        -------
        Dict[str, Any]
            A dictionary containing the explanation and additional information.
            Namely, the following keys are present:
            - "prediction": The prediction of the model.
            - "explained_class_idx": The class (index) of the explained output.
            - "dynamic_linear_weights": The dynamic linear weights of the model (`in_tensor.grad`).
            - "contribution_map": The contribution map of the model prediction.
            - "explanation": The explanation of the model prediction.
        """
        if in_tensor.ndim == 3:
            raise ValueError("Expected 4-dimensional input tensor")
        if in_tensor.shape[0] != 1:
            raise ValueError("Expected batch size of 1")
        if not in_tensor.requires_grad:
            warnings.warn(
                "Input tensor did not require grad! Has been set automatically to True!"
            )
            in_tensor.requires_grad = True  # nonsense otherwise
        if self.training:  # noqa
            warnings.warn(
                "Model is in training mode! "
                "This might lead to unexpected results! Use model.eval()!"
            )

        result = dict()
        with torch.enable_grad(), self.explanation_mode():
            # fwd + prediction
            out = self(in_tensor)  # noqa
            pred_out = out.max(1)
            result["prediction"] = pred_out.indices.item()

            # select output (logit) to explain
            if idx is None:  # explain prediction
                to_be_explained_logit = pred_out.values
                result["explained_class_idx"] = pred_out.indices.item()
            else:  # user specified idx
                to_be_explained_logit = out[0, idx]
                result["explained_class_idx"] = idx

            to_be_explained_logit.backward(inputs=[in_tensor])

        # get weights and contribution map
        result["dynamic_linear_weights"] = in_tensor.grad
        result["contribution_map"] = (in_tensor * in_tensor.grad).sum(1)

        # generate (color) explanation
        result["explanation"] = gradient_to_image(
            in_tensor[0], in_tensor.grad[0], **grad2img_kwargs
        )

        return result

    @staticmethod  # to make it easier when using torch.hub
    def gradient_to_image(
        image: "Tensor",
        linear_mapping: "Tensor",
        smooth: int = 15,
        alpha_percentile: float = 99.5,
    ) -> "np.ndarray":
        """
        From https://github.com/moboehle/B-cos/blob/0023500ce/interpretability/utils.py#L41.
        Computing color image from dynamic linear mapping of B-cos models.

        Parameters
        ----------
        image: Tensor
            Original input image (encoded with 6 color channels)
            Shape: [C, H, W] with C=6
        linear_mapping: Tensor
            Linear mapping W_{1\rightarrow l} of the B-cos model
            Shape: [C, H, W] same as image
        smooth: int
            Kernel size for smoothing the alpha values
        alpha_percentile: float
            Cut-off percentile for the alpha value

        Returns
        -------
        np.ndarray
            image explanation of the B-cos model.
            Shape: [H, W, C] (C=4 ie RGBA)
        """
        return gradient_to_image(image, linear_mapping, smooth, alpha_percentile)

    @staticmethod  # to make it easier when using torch.hub
    def plot_contribution_map(
        contribution_map: TensorLike,
        ax: Optional["plt.Axes"] = None,
        vrange: Optional[float] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        hide_ticks: bool = True,
        cmap: str = "bwr",
        percentile: float = 99.5,
    ) -> "Tuple[plt.Axes, matplotlib.image.AxesImage]":
        """
        From https://github.com/moboehle/B-cos/blob/0023500cea7b/interpretability/utils.py#L78-L115
        For an example of how to use this, see docstring for `explain`.

        Visualises a contribution map, i.e., a matrix assigning individual weights to each spatial location.
        As default, this shows a contribution map with the "bwr" colormap and chooses vmin and vmax so that the map
        ranges from (-max(abs(contribution_map), max(abs(contribution_map)).

        Parameters
        ----------
        contribution_map: TensorLike
            (H, W) matrix of contributions to visualize.
        ax: Optional[plt.Axes]
            Axis on which to plot. If None, a new figure is created.
        vrange: Optional[float]
            If None, the colormap ranges from -v to v, with v being the maximum absolute value in the map.
            If provided, it will range from -vrange to vrange, as long as either one of the boundaries is not
            overwritten by vmin or vmax.
        vmin: Optional[float]
            Manually overwrite the minimum value for the colormap range instead of using -vrange.
        vmax: Optional[float]
            Manually overwrite the maximum value for the colormap range instead of using vrange.
        hide_ticks: bool
            Sets the axis ticks to []
        cmap: str
            colormap to use for the contribution map plot.
        percentile: float
            If percentile is given, this will be used as a cut-off for the attribution maps.

        Returns
        -------
        ax: plt.Axes
            The axis on which the contribution map was plotted.
        im: matplotlib.image.AxesImage
            The image object of the contribution map.
        """
        return plot_contribution_map(
            contribution_map,
            ax,
            vrange,
            vmin,
            vmax,
            hide_ticks,
            cmap,
            percentile,
        )

    def attribute(
        self,
        image: Union[Tensor, Tuple[Tensor]],
        target: Union[int, Tuple[int], Tensor, List[int]],
        **kwargs: Any,
    ) -> Tensor:
        """
        From https://github.com/moboehle/B-cos/blob/4cd3b8ffc24d64c8b5b3262479bd/training/trainer_base.py#L447-L464
        This is essentially just a Captum-IxG-dependent contribution generator for B-cos models.

        This method returns the contribution map according to Input x Gradient.
        Specifically, if the prediction model is a dynamic linear network, it returns the contribution map according
        to the linear mapping (IxG with detached dynamic weights).

        Parameters
        ----------
        image: Tensor | Tuple[Tensor]
            Input image(s). Shape (1, C, H, W).
        target: int | tuple[int] | Tensor | list[int]
            Target class to check contributions for.
        kwargs: Any
            just for compatibility...

        Returns
        -------
        Tensor
            Contributions for desired level.
        """
        _ = kwargs
        # install captum if you want to use captum dependent contribution generation
        # this will fail otherwise
        from interpretability.explanation_methods.explainers.captum import IxG

        with self.explanation_mode():
            attribution_f = IxG(self)
            att = attribution_f.attribute(image, target)

        return att

    def attribute_selection(
        self,
        image: Tensor,
        targets: Union[Tuple[int], Tensor, List[int]],
        **kwargs: Any,
    ) -> Tensor:
        """
        From https://github.com/moboehle/B-cos/blob/4cd3b8ffc24d64c8b5b3262479/training/trainer_base.py#L467-L481
        Runs `.attribute` for multiple targets and concatenates the results.

        Parameters
        ----------
        image: Tensor
            Input image.
        targets: tuple[int] | Tensor | list[int]
            Target classes to check contributions for.
        kwargs: Any
            just for compatibility...

        Returns
        -------
        Tensor
            Contributions for desired level.
        """
        _ = kwargs
        return torch.cat([self.attribute(image, t) for t in targets], dim=0)


class explanation_mode(_DecoratorContextManager):
    """
    A context manager which activates and puts model in to explanation
    mode and deactivates it afterwards.
    Can also be used as a decorator.

    Parameters
    ----------
    model : nn.Module
        The model to put in explanation mode.
    """

    def __init__(self, model: "nn.Module"):
        self.model = model
        self.expl_modules = None

    def find_expl_modules(self) -> None:
        """Finds all modules which have a `set_explanation_mode` method."""
        self.expl_modules = [
            m for m in self.model.modules() if hasattr(m, "set_explanation_mode")
        ]

    def __enter__(self):
        """
        Put model in explanation mode.
        """
        if self.expl_modules is None:
            self.find_expl_modules()

        for m in self.expl_modules:
            m.set_explanation_mode(True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Turn off explanation mode for model.
        """
        for m in self.expl_modules:
            m.set_explanation_mode(False)


def gradient_to_image(image, linear_mapping, smooth=15, alpha_percentile=99.5):
    """
    From https://github.com/moboehle/B-cos/blob/0023500ce/interpretability/utils.py#L41.
    Computing color image from dynamic linear mapping of B-cos models.

    Parameters
    ----------
    image: Tensor
        Original input image (encoded with 6 color channels)
        Shape: [C, H, W] with C=6
    linear_mapping: Tensor
        Linear mapping W_{1\rightarrow l} of the B-cos model
        Shape: [C, H, W] same as image
    smooth: int
        Kernel size for smoothing the alpha values
    alpha_percentile: float
        Cut-off percentile for the alpha value. In range [0, 100].

    Returns
    -------
    np.ndarray
        image explanation of the B-cos model.
        Shape: [H, W, C] (C=4 ie RGBA)
    """
    # shape of img and linmap is [C, H, W], summing over first dimension gives the contribution map per location
    contribs = (image * linear_mapping).sum(0, keepdim=True)  # [H, W]
    # Normalise each pixel vector (r, g, b, 1-r, 1-g, 1-b) s.t. max entry is 1, maintaining direction
    rgb_grad = linear_mapping / (
        linear_mapping.abs().max(0, keepdim=True).values + 1e-12
    )
    # clip off values below 0 (i.e., set negatively weighted channels to 0 weighting)
    rgb_grad = rgb_grad.clamp(min=0)
    # normalise s.t. each pair (e.g., r and 1-r) sums to 1 and only use resulting rgb values
    rgb_grad = rgb_grad[:3] / (rgb_grad[:3] + rgb_grad[3:] + 1e-12)  # [3, H, W]

    # Set alpha value to the strength (L2 norm) of each location's gradient
    alpha = linear_mapping.norm(p=2, dim=0, keepdim=True)
    # Only show positive contributions
    alpha = torch.where(contribs < 0, 1e-12, alpha)
    if smooth:
        alpha = F.avg_pool2d(alpha, smooth, stride=1, padding=(smooth - 1) // 2)
    alpha = (alpha / torch.quantile(alpha, q=alpha_percentile / 100)).clip(0, 1)

    rgb_grad = torch.concatenate([rgb_grad, alpha], dim=0)  # [4, H, W]
    # Reshaping to [H, W, C]
    grad_image = rgb_grad.permute(1, 2, 0)
    return grad_image.detach().cpu().numpy()


def plot_contribution_map(
    contribution_map: TensorLike,
    ax: Optional["plt.Axes"] = None,
    vrange: Optional[float] = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    hide_ticks: bool = True,
    cmap: str = "bwr",
    percentile: float = 99.5,
) -> "Tuple[plt.Axes, matplotlib.image.AxesImage]":
    """
    From https://github.com/moboehle/B-cos/blob/0023500cea7b/interpretability/utils.py#L78-L115
    For an example of how to use this, see docstring for `BcosMixin.explain`.

    Visualises a contribution map, i.e., a matrix assigning individual weights to each spatial location.
    As default, this shows a contribution map with the "bwr" colormap and chooses vmin and vmax so that the map
    ranges from (-max(abs(contribution_map), max(abs(contribution_map)).

    Parameters
    ----------
    contribution_map: TensorLike
        (H, W) matrix of contributions to visualize.
    ax: Optional[plt.Axes]
        Axis on which to plot. If None, a new figure is created.
    vrange: Optional[float]
        If None, the colormap ranges from -v to v, with v being the maximum absolute value in the map.
        If provided, it will range from -vrange to vrange, as long as either one of the boundaries is not
        overwritten by vmin or vmax.
    vmin: Optional[float]
        Manually overwrite the minimum value for the colormap range instead of using -vrange.
    vmax: Optional[float]
        Manually overwrite the maximum value for the colormap range instead of using vrange.
    hide_ticks: bool
        Sets the axis ticks to []
    cmap: str
        colormap to use for the contribution map plot.
    percentile: float
        If percentile is given, this will be used as a cut-off for the attribution maps.

    Returns
    -------
    ax: plt.Axes
        The axis on which the contribution map was plotted.
    im: matplotlib.image.AxesImage
        The image object of the contribution map.
    """
    assert (
        contribution_map.ndim == 2
    ), "Contribution map is supposed to only have 2 spatial dimensions."
    if isinstance(contribution_map, torch.Tensor):
        contribution_map = contribution_map.detach().cpu().numpy()
    cutoff = np.percentile(np.abs(contribution_map), percentile)
    contribution_map = np.clip(contribution_map, -cutoff, cutoff)

    if ax is None:
        import matplotlib.pyplot as plt  # noqa

        fig, ax = plt.subplots(1)

    if vrange is None or vrange == "auto":
        vrange = np.max(np.abs(contribution_map.flatten()))
    im = ax.imshow(
        contribution_map,
        cmap=cmap,
        vmin=-vrange if vmin is None else vmin,
        vmax=vrange if vmax is None else vmax,
    )

    if hide_ticks:
        ax.set_xticks([])
        ax.set_yticks([])

    return ax, im



# From https://github.com/BaguaSys/bagua
import os
import socket


def get_world_size() -> int:
    """
    Get the number of processes in the default process group.

    Returns:
        The world size of the default process group.
    """
    return int(os.environ.get("WORLD_SIZE", 1))


def get_rank() -> int:
    """
    Get the rank of the default process group.

    Rank is a unique identifier assigned to each process within the default
    process group. They are always consecutive integers ranging from 0 to
    ``world_size``.

    Returns:
        The rank of the default process group.
    """
    return int(os.environ.get("RANK", 0))


def get_local_rank() -> int:
    """
    Get the rank of current node.

    Local rank is a unique identifier assigned to each process within a node.
    They are always consecutive integers ranging from 0 to ``local_size``.

    Returns:
        The local rank of the node.
    """
    return int(os.environ.get("LOCAL_RANK", 0))


def get_local_size() -> int:
    """
    Get the number of processes in the node.

    Returns:
        The local size of the node.
    """
    return int(os.environ.get("LOCAL_WORLD_SIZE", 1))


def get_node_rank() -> int:
    """
    Get the rank among all nodes.

    Returns:
        The node rank of the node.
    """
    if _is_elastic_launched():
        return int(os.environ.get("GROUP_RANK", 0))
    else:
        if "SLURM_NODEID" in os.environ:
            return int(os.environ["SLURM_NODEID"])
        else:
            return int(os.environ.get("NODE_RANK", 0))


def _is_elastic_launched():
    """Returns ``True`` if the current process was launched using the bagua.distributed.run command."""
    required_env_vars = {"RANK", "GROUP_RANK", "LOCAL_RANK", "LOCAL_WORLD_SIZE"}
    return required_env_vars.issubset(os.environ.keys())


def get_master_addr() -> str:
    return os.environ.get("MASTER_ADDR", "127.0.0.1")


def find_free_network_port() -> int:
    """Finds a free port on localhost."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port
